process_colors = {
    "Data": "black",
    "sig": "orange",
    "bkg": "dodgerblue",
    "bkg_2": "turquoise",
}


process_labels = {
    "Data": "Data",
    "sig": r"$X^{\psi}\to\rho\lambda$",
    "bkg": "Bkg.",
    "bkg_2": "M[0]",
}


axis_labels = {
    "a": r"Variable a",
    "b": {"label": r"Variable b", "unit": "GeV"},
    "x": {"label": r"$\mathit{p}_{T}^{\mu+MET}$", "unit": "GeV"},
}


# for impact plots, must be in html style
impact_labels = {
    "slope_signal": "Signal slope",
    "slope_signal_ch0": "Slope signal (ch0)",
    "slope_signal_ch1": "Slope signal (ch1)",
    "slope_2_signal_ch1": "Slope signal, 2 (ch1)",
    "slope_background": "Bkg. signal",
    "slope_quad_signal_ch0SymDiff": "slope <i>μ</i><sup>2</sup> [diff]",
    "slope_quad_signal_ch0SymAvg": "slope <i>μ</i><sup>2</sup> [avg]",
    "slope_lin_signal_ch0SymDiff": "slope <i>x</i><sup>1</sup> [avg]",
    "slope_lin_signal_ch0SymAvg": "slope <i>x</i><sup>1</sup> [diff]",
    "norm": "Normalization",
    "bkg_norm": "Bkg. normalization",
    "bkg_2_norm": "Bkg. (2) normalization",
    "Total": "All",
    "stat": "Data stat.",
    "binByBinStat": "Simulated samples stat.",
    "slopes": "Slopes",
    "slopes_signal": "Signal slopes",
    "slopes_background": "Background slopes",
    "sig": "Signal",
}


def translate_html_to_latex(n):
    # transform html style formatting into latex style
    if "</" in n:
        n = (
            f"${n}$".replace("<i>", r"\mathit{")
            .replace("<sub>", "_{")
            .replace("<sup>", "^{")
            .replace("</i>", "}")
            .replace("</sub>", "}")
            .replace("</sup>", "}")
            .replace(" ", r"\ ")
        )
    return n


# same as impact labels but in latex format
systematics_labels = {k: translate_html_to_latex(v) for k, v in impact_labels.items()}

# predefined set of grouped impacts to be plotted
nuisance_grouping = {
    "min": ["Total", "stat", "binByBinStat", "slopes", "norm"],
}



#!/usr/bin/env python3

import argparse
import os
from pprint import pprint

from wums import logging

from rabbit.datacard_converter import DatacardConverter

logger = None


def main():
    parser = argparse.ArgumentParser(
        description="Convert Combine datacard and ROOT files to different formats"
    )
    parser.add_argument("datacard", help="Path to the datacard file")
    parser.add_argument(
        "-o",
        "--output",
        default=None,
        help="output directory, if 'None' same as input datacard",
    )
    parser.add_argument(
        "--outname",
        default=None,
        help="output file name, if 'None' same as input datacard but with .hdf5 extension",
    )
    parser.add_argument(
        "--postfix",
        default=None,
        type=str,
        help="Postfix to append on output file name",
    )
    parser.add_argument(
        "--sparse",
        default=False,
        action="store_true",
        help="Make sparse tensor",
    )
    parser.add_argument(
        "--symmetrize",
        default=None,
        choices=[None, "conservative", "average", "linear", "quadratic"],
        type=str,
        help="Symmetrize tensor by forcing systematics to 'average'",
    )
    parser.add_argument(
        "--mass",
        type=str,
        default="125.38",
        help="Higgs boson mass to replace $MASS string in datacard",
    )
    parser.add_argument(
        "--root",
        action="store_true",
        help="Use root to load histograms, otherwise uproot",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        type=int,
        default=3,
        choices=[0, 1, 2, 3, 4],
        help="Set verbosity level with logging, the larger the more verbose",
    )
    parser.add_argument(
        "--noColorLogger", action="store_true", help="Do not use logging with colors"
    )

    args = parser.parse_args()

    global logger
    logger = logging.setup_logger(__file__, args.verbose, args.noColorLogger)

    converter = DatacardConverter(
        args.datacard, use_root=args.root, mass=args.mass, symmetrize=args.symmetrize
    )
    writer = converter.convert_to_hdf5(sparse=args.sparse)

    pprint(converter.parser.get_summary())

    directory = args.output
    if directory is None:
        directory = os.path.dirname(args.datacard)
    if directory == "":
        directory = "./"
    filename = args.outname
    if filename is None:
        filename = os.path.splitext(os.path.basename(args.datacard))[0]
    if args.postfix:
        filename += f"_{args.postfix}"
    writer.write(outfolder=directory, outfilename=filename)

    del converter


if __name__ == "__main__":
    main()



import tensorflow as tf

from rabbit.physicsmodels.physicsmodel import Channelmodel


class Project(Channelmodel):
    """
    A class to project a histogram to lower dimensions.
    The normalization is done to the integral of all processes or data.

    Parameters
    ----------
    channel_name : str
        Name of the channel. Required.
    axes_names : list of str, optional
        Names of the axes to keep. If empty, the histogram will be projected to a single bin.
    """

    def __init__(self, indata, key, channel, *axes_names):
        super().__init__(indata, key, channel)
        self.channel = channel

        info = indata.channel_info[channel]
        channel_axes = {a.name: a for a in info["axes"]}

        if len([n for n in axes_names if n not in channel_axes]) > 0:
            raise RuntimeError(
                f"Axes {[n for n in axes_names if n not in channel_axes]} not found. Available axes are {[channel_axes.keys()]}"
            )
        hist_axes = [channel_axes[n] for n in axes_names]

        if len(hist_axes) != len(axes_names):
            raise ValueError(
                f"Hist axes {[h.name for h in hist_axes]} != {axes_names} not found"
            )

        channel_axes_names = [axis.name for axis in channel_axes.values()]

        axis_idxs = [channel_axes_names.index(axis) for axis in axes_names]

        self.proj_idxs = [i for i in range(len(channel_axes)) if i not in axis_idxs]

        post_proj_axes_names = [
            axis for axis in channel_axes_names if axis in axes_names
        ]

        self.transpose_idxs = [post_proj_axes_names.index(axis) for axis in axes_names]

        self.has_data = not info.get("masked", False)

        self.channel_info = {
            channel: {
                "axes": hist_axes,
                "flow": info.get("flow", False),
                "processes": indata.procs,
            }
        }

    def project(self, values):
        exp = tf.reduce_sum(values, axis=self.proj_idxs)
        perm = self.transpose_idxs[:]
        if len(exp.shape) > len(self.transpose_idxs):
            # last is process axis
            perm += list(range(len(self.transpose_idxs), len(exp.shape)))
        exp = tf.transpose(exp, perm=perm)
        return exp

    def compute(self, param, observables):
        return self.project(observables)


class Normalize(Project):
    """
    Same as project but also normalize
    """

    ndf_reduction = 1

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def project(self, values, *args):
        norm = tf.reduce_sum(values)
        out = values / norm
        return super().project(out, *args)



#!/usr/bin/env python3

import argparse
import os

import mplhep as hep
import numpy as np
from wums import logging

from rabbit import io_tools

from wums import logging, output_tools, plot_tools  # isort: skip

hep.style.use(hep.style.ROOT)


logger = None


def parseArgs():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-v",
        "--verbose",
        type=int,
        default=3,
        choices=[0, 1, 2, 3, 4],
        help="Set verbosity level with logging, the larger the more verbose",
    )
    parser.add_argument(
        "--noColorLogger", action="store_true", help="Do not use logging with colors"
    )
    parser.add_argument(
        "infiles",
        type=str,
        nargs="*",
        help="hdf5 files from rabbit or root file from combinetf",
    )
    parser.add_argument(
        "-o",
        "--outpath",
        type=str,
        default=os.path.expanduser("./test"),
        help="Base path for output",
    )
    parser.add_argument(
        "--noObs", action="store_true", help="Don't plot observed limit"
    )
    parser.add_argument(
        "--xvals", type=float, nargs="+", default=None, help="x-axis ticks for plotting"
    )
    parser.add_argument(
        "--biases",
        type=float,
        nargs="+",
        default=None,
        help="Add additional line to highlight the bias in each point",
    )
    parser.add_argument(
        "--scale", type=float, default=1.0, help="Scale limits by this value"
    )
    parser.add_argument(
        "--config",
        type=str,
        default=None,
        help="Path to config file for style formatting",
    )
    parser.add_argument(
        "--eoscp",
        action="store_true",
        help="Override use of xrdcp and use the mount instead",
    )
    parser.add_argument(
        "-p", "--postfix", type=str, help="Postfix for output file name"
    )
    parser.add_argument(
        "--xlabel", type=str, default=None, help="x-axis label for plot labeling"
    )
    parser.add_argument(
        "--ylabel", type=str, default=None, help="y-axis label for plot labeling"
    )
    parser.add_argument(
        "--title",
        default="Rabbit",
        type=str,
        help="Title to be printed in upper left",
    )
    parser.add_argument(
        "--subtitle",
        default="",
        type=str,
        help="Subtitle to be printed after title",
    )
    parser.add_argument("--titlePos", type=int, default=2, help="title position")
    parser.add_argument(
        "--legPos", type=str, default="upper right", help="Set legend position"
    )
    parser.add_argument(
        "--legSize",
        type=str,
        default="small",
        help="Legend text size (small: axis ticks size, large: axis label size, number)",
    )
    parser.add_argument(
        "--legCols", type=int, default=2, help="Number of columns in legend"
    )
    parser.add_argument(
        "--ylim",
        type=float,
        nargs=2,
        default=None,
        help="Min and max values for y axis (if not specified, range set automatically)",
    )
    args = parser.parse_args()

    return args


def main():
    """
    Plot the covariance matrix of the histogram bins
    """

    args = parseArgs()

    outdir = output_tools.make_plot_dir(args.outpath, eoscp=args.eoscp)

    logger = logging.setup_logger(__file__, args.verbose, args.noColorLogger)

    cls_list = set()
    clb_list = set()
    limits = []
    limits_asimov = []
    lumi = None
    for infile in args.infiles:
        fitresult_asimov, meta = io_tools.get_fitresult(
            infile, result="asimov", meta=True
        )

        h_limits_asimov = fitresult_asimov["asymptoticLimits"].get()

        cls_list = cls_list.union(set([cls for cls in h_limits_asimov.axes["cls"]]))
        clb_list = clb_list.union(set([clb for clb in h_limits_asimov.axes["clb"]]))

        limits_asimov.append(h_limits_asimov)

        fitresult, meta = io_tools.get_fitresult(infile, meta=True)
        if fitresult != fitresult_asimov and not args.noObs:
            if (
                "asymptoticLimits" in fitresult.keys()
                and "clb" not in fitresult["asymptoticLimits"].get().axes.name
            ):
                h_limits = fitresult["asymptoticLimits"].get()
                cls_list = cls_list.union(set([cls for cls in h_limits.axes["cls"]]))
                limits.append(h_limits)
            else:
                limits.append(h_limits_asimov[{"clb": "0.5"}])

        channel_info = meta["meta_info_input"]["channel_info"]
        lumis = set([c.get("lumi", None) for c in channel_info.values()])
        if len(lumis) > 1 or (lumi is not None and tuple(lumis)[0] != lumi):
            raise ValueError("Different channels with different luminosity values")
        lumi = tuple(lumis)[0]

    if args.xvals is not None:
        x = np.array(args.xvals)
    else:
        x = np.arange(len(limits_asimov))

    clb_list = list(clb_list)

    for cls in cls_list:

        yexp = np.array(
            [l[{"cls": cls, "clb": "0.5"}] * args.scale for l in limits_asimov]
        ).flatten()

        yexp_m2 = np.array(
            [l[{"cls": cls, "clb": "0.025"}] * args.scale for l in limits_asimov]
        ).flatten()
        yexp_m1 = np.array(
            [l[{"cls": cls, "clb": "0.16"}] * args.scale for l in limits_asimov]
        ).flatten()
        yexp_p1 = np.array(
            [l[{"cls": cls, "clb": "0.84"}] * args.scale for l in limits_asimov]
        ).flatten()
        yexp_p2 = np.array(
            [l[{"cls": cls, "clb": "0.975"}] * args.scale for l in limits_asimov]
        ).flatten()

        ylist = [yexp, yexp_m1, yexp_m2, yexp_p1, yexp_p2]

        if len(limits) > 0:
            yobs = np.array([l[{"cls": cls}] * args.scale for l in limits]).flatten()
            ylist.append(yobs)

        ylist = np.array(ylist)
        ymin, ymax = np.min(ylist), np.max(ylist)
        yrange = ymax - ymin

        fig, ax = plot_tools.figure(
            None,
            args.xlabel,
            args.ylabel,
            automatic_scale=False,
            xlim=(min(x) - 0.5, max(x) + 0.5),
            ylim=(
                (ymin - yrange * 0.1, ymax + yrange * 0.6)
                if args.ylim is None
                else args.ylim
            ),
        )

        ax.fill_between(x, yexp_m2, yexp_p2, color="#F5BB54", label=r"95% expected")

        ax.fill_between(x, yexp_m1, yexp_p1, color="#607641", label=r"68% expected")

        # Expected (median)
        ax.plot(
            x,
            yexp,
            linestyle="--",
            color="black",
            linewidth=1.5,
            label="Median expected",
        )

        if args.biases is not None:
            color_bias = "tab:red"
            # add right y-axis
            ax2 = ax.twinx()
            ax2.plot(
                x,
                args.biases,
                linestyle="--",
                color=color_bias,
                linewidth=1.5,
                label="Expected bias",
            )
            ax2.set_ylabel("1 - relative expected bias", color=color_bias)
            ax2.tick_params(axis="y", labelcolor=color_bias)

        if len(limits) > 0:
            # Observed
            ax.plot(x, yobs, color="black", linewidth=1.5, label="Observed")

        plot_tools.add_decor(
            ax,
            args.title,
            args.subtitle,
            data=len(limits) > 0,
            lumi=lumi,  # if args.dataName == "Data" and not args.noData else None,
            loc=args.titlePos,
            text_size=args.legSize,
        )
        plot_tools.addLegend(
            ax,
            text_size=args.legSize,
            ncols=args.legCols,
            loc=args.legPos,
            title=f"{round((1-float(cls)) * 100)}% CL upper limits",
        )

        to_join = ["limits", args.postfix, f"CLs{str(cls).replace('.','p')}"]
        outfile = "_".join(filter(lambda x: x, to_join))
        if args.subtitle == "Preliminary":
            outfile += "_preliminary"

        plot_tools.save_pdf_and_png(outdir, outfile)

        analysis_meta_info = None
        if meta is not None:
            if "meta_info_input" in meta:
                analysis_meta_info = {
                    "RabbitOutput": meta["meta_info"],
                    "AnalysisOutput": meta["meta_info_input"]["meta_info"],
                }
            else:
                analysis_meta_info = {"AnalysisOutput": meta["meta_info"]}

        output_tools.write_index_and_log(
            outdir,
            outfile,
            analysis_meta_info={
                # "Stacked processes": yield_tables["stacked"],
                # "Unstacked processes": yield_tables["unstacked"],
                **analysis_meta_info,
            },
            args=args,
        )

    if output_tools.is_eosuser_path(args.outpath) and args.eoscp:
        output_tools.copy_to_eos(outdir, args.outpath)


if __name__ == "__main__":
    main()



import tensorflow as tf

from rabbit.physicsmodels import helpers
from rabbit.physicsmodels.physicsmodel import PhysicsModel


class AngularCoefficients(PhysicsModel):
    """
    A class to compute the angular coefficients as A_i = sigma_i / sigma_UL

    Parameters
    ----------
        indata: Input data used for analysis (e.g., histograms or data structures).
        channel: str
            Name of the channel.
        process: str
            Name of the process.
        selection: dict, optional
            Dictionary specifying selection criteria. Keys are axis names, and values are slices or conditions.
            Defaults to an empty dictionary meaning no selection.
            E.g. {"charge":0, "ptVgen":slice(0,2), "absYVgen": hist.sum, "massVgen": hist.rebin(2)}
            Selected axes are summed before the ratio is computed. To integrate over one axis before the ratio, use `slice(None)`
    """

    def __init__(
        self,
        indata,
        key,
        channel,
        processes=[],
        selection={},
        rebin_axes={},
        sum_axes=[],
        helicity_axis=None,
    ):
        self.key = key

        if helicity_axis is None:
            axes_names = [a.name for a in indata.channel_info[channel]["axes"]]
            if "helicitySig" in axes_names:
                helicity_axis = "helicitySig"
            elif "helicity" in axes_names:
                helicity_axis = "helicity"
            else:
                raise RuntimeError(
                    f"Could not find helicity axes, available axes are {axes_names}"
                )

        # sigma_i
        self.num = helpers.Term(
            indata,
            channel,
            processes,
            {helicity_axis: slice(1, None), **selection},
            rebin_axes,
            sum_axes,
        )
        # sigma_UL
        self.den = helpers.Term(
            indata,
            channel,
            processes,
            {helicity_axis: 0, **selection},
            rebin_axes,
            sum_axes,
        )

        self.has_data = False

        self.need_processes = False

        # The output of ratios will always be without process axis
        self.skip_per_process = True

        self.has_processes = False  # The result has no process axis

        self.helicity_index = [
            i for i, a in enumerate(self.num.channel_axes) if a.name == helicity_axis
        ][0]

        channel_axes = []
        for a in self.num.channel_axes:
            if a.name == helicity_axis:
                a.__dict__["name"] = "ai"
            channel_axes.append(a)

        self.channel_info = {
            channel: {
                "axes": channel_axes,
                "flow": indata.channel_info[channel].get("flow", False),
            },
        }

    @classmethod
    def parse_args(cls, indata, *args):
        """
        parsing the input arguments into the AI constructor, it has to be called as
        -m AngularCoefficients
            <ch>
            <proc_0>,<proc_1>,...
            <axis_ai_0>:<slice_ai_0>,<axis_ai_1>,<slice_ai_1>...

        Processes selections are optional. Use 'None' if you don't want to select any.
        Axes selections are optional.
        """

        if len(args) > 2 and ":" not in args[1]:
            procs = [p for p in args[1].split(",") if p != "None"]
        else:
            procs = []

        # find axis selections
        if any(a for a in args if ":" in a):
            sel_args = [a for a in args if ":" in a]
            axis_selection, axes_rebin, axes_sum = helpers.parse_axis_selection(
                sel_args[0]
            )
        else:
            axis_selection = {}
            axes_sum = []
            axes_rebin = {}

        key = " ".join([cls.__name__, *args])

        return cls(
            indata,
            key,
            args[0],
            procs,
            axis_selection,
            axes_rebin,
            axes_sum,
        )

    def compute_ais(self, observables, inclusive=False):
        num = self.num.select(observables, inclusive=inclusive)
        den = self.den.select(observables, inclusive=inclusive)
        den = tf.expand_dims(den, axis=self.helicity_index)
        return num / den

    def compute_flat(self, params, observables):
        exp = self.compute_ais(observables, True)
        return tf.reshape(exp, [-1])

    def compute_flat_per_process(self, params, observables):
        exp = self.compute_ais(observables, False)
        flat_shape = (-1, exp.shape[-1])
        return tf.reshape(exp, [-1, flat_shape])


class LamTung(AngularCoefficients):

    def __init__(self, indata, key, channel, *args, **kwargs):
        super().__init__(indata, key, channel, *args, **kwargs)

        self.channel_info[channel] = {
            "axes": [c for c in self.channel_info[channel]["axes"] if c.name != "ai"],
            "flow": self.channel_info[channel].get("flow", False),
        }

    def compute_ais(self, observables, inclusive=False):
        ais = super().compute_ais(observables, inclusive)
        a0 = tf.gather(ais, indices=0, axis=self.helicity_index)
        a2 = tf.gather(ais, indices=2, axis=self.helicity_index)
        return a0 - a2



import math

import numpy as np
import tensorflow as tf


def maketensor(h5dset):
    if "original_shape" in h5dset.attrs:
        shape = h5dset.attrs["original_shape"]
    else:
        shape = h5dset.shape

    if h5dset.size == 0:
        return tf.zeros(shape, h5dset.dtype)

    # read directly from hdf5 dataset to the underlying buffer of a tensor
    # this requires that the tensor is located on the CPU, so force the device
    with tf.device(tf.config.list_logical_devices("CPU")[0]):
        atensor = tf.zeros(h5dset.shape, h5dset.dtype)
        # zero tensors have a special flag set, using the identity clears this implicitly
        atensor = tf.identity(atensor)

    # read into the underlying array
    h5dset.read_direct(atensor.__array__())

    # the reshape operation is needed in case the hdf5 dataset was flattened
    # this also triggers a copy of the tensor to the default device (e.g. GPU)
    # if needed (ie if not the default CPU device)
    atensor = tf.reshape(atensor, shape)
    return atensor


def makesparsetensor(h5group):
    indices = maketensor(h5group["indices"])
    values = maketensor(h5group["values"])
    dense_shape = h5group.attrs["dense_shape"]

    return tf.sparse.SparseTensor(indices, values, dense_shape)


def writeFlatInChunks(arr, h5group, outname, maxChunkBytes=1024**2):
    arrflat = arr.reshape(-1)

    esize = np.dtype(arrflat.dtype).itemsize
    nbytes = arrflat.size * esize

    # special handling for empty datasets, which should not use chunked storage or compression
    if arrflat.size == 0:
        chunksize = 1
        chunks = None
        compression = None
    else:
        chunksize = int(min(arrflat.size, max(1, math.floor(maxChunkBytes / esize))))
        chunks = (chunksize,)
        compression = "gzip"

    h5dset = h5group.create_dataset(
        outname,
        arrflat.shape,
        chunks=chunks,
        dtype=arrflat.dtype,
        compression=compression,
    )

    # write in chunks, preserving sparsity if relevant
    for ielem in range(0, arrflat.size, chunksize):
        aout = arrflat[ielem : ielem + chunksize]
        if np.count_nonzero(aout):
            h5dset[ielem : ielem + chunksize] = aout

    h5dset.attrs["original_shape"] = np.array(arr.shape, dtype="int64")

    return nbytes


def writeSparse(indices, values, dense_shape, h5group, outname, maxChunkBytes=1024**2):
    outgroup = h5group.create_group(outname)

    nbytes = 0
    nbytes += writeFlatInChunks(indices, outgroup, "indices", maxChunkBytes)
    nbytes += writeFlatInChunks(values, outgroup, "values", maxChunkBytes)
    outgroup.attrs["dense_shape"] = np.array(dense_shape, dtype="int64")

    return nbytes



#!/usr/bin/env python3

import argparse
import itertools
import math
import os
import re

import numpy as np
import pandas as pd
import plotly.graph_objects as go
import plotly.io as pio
from plotly.subplots import make_subplots

from rabbit import io_tools

from wums import output_tools, plot_tools  # isort: skip


# prevent MathJax from bein loaded
pio.kaleido.scope.mathjax = None


def writeOutput(fig, outfile, extensions=[], postfix=None, args=None, meta_info=None):
    name, _ = os.path.splitext(outfile)

    if postfix:
        name += f"_{postfix}"

    for ext in extensions:
        if ext[0] != ".":
            ext = "." + ext
        output = name + ext
        print(f"Write output file {output}")
        if ext == ".html":
            fig.write_html(output, include_mathjax=False)
        else:
            fig.write_image(output)

        output = name.rsplit("/", 1)
        output[1] = os.path.splitext(output[1])[0]
        if len(output) == 1:
            output = (None, *output)
    if args is None and meta_info is None:
        return
    output_tools.write_logfile(
        *output,
        args=args,
        meta_info=meta_info,
    )


def get_marker(filled=True, color="#377eb8", opacity=1.0):
    if filled:
        marker = {
            "marker": {
                "color": color,  # Fill color for the filled bars
                "opacity": opacity,  # Opacity for the filled bars (adjust as needed)
            }
        }
    else:
        marker = {
            "marker": {
                "color": "rgba(0, 0, 0, 0)",  # Transparent fill color
                "opacity": opacity,
                "line": {"color": color, "width": 2},  # Border color  # Border width
            }
        }
    return marker


def plotImpacts(
    df,
    impact_title="Impacts",
    pulls=False,
    oneSidedImpacts=False,
    pullrange=None,
    title=None,
    subtitle=None,
    impacts=True,
    asym=False,
    asym_pulls=False,
    include_ref=False,
    ref_name="ref.",
    name="",
    show_numbers=False,
    show_legend=True,
    legend_pos="bottom",
    group=None,
    diff_pulls=True,
):
    impacts = impacts and bool(np.count_nonzero(df["absimpact"]))
    ncols = pulls + impacts
    fig = make_subplots(rows=1, cols=ncols, horizontal_spacing=0.1, shared_yaxes=True)

    loffset = 40
    if title is not None:
        if subtitle is not None:
            loffset += max(len(subtitle), len(title)) * 7
        else:
            loffset += len(title) * 6

    if legend_pos == "bottom":
        legend = dict(
            orientation="h",
            xanchor="left",
            yanchor="top",
            x=0.0,
            y=0.0,
        )
    elif legend_pos == "right":
        legend = dict(
            orientation="v",
            xanchor="left",
            yanchor="top",
            x=1.0,
            y=1.0,
        )
    else:
        raise NotImplementedError("Supported legend positions are ['bottom', 'left']")

    ndisplay = len(df)
    fig.update_layout(
        paper_bgcolor="rgba(100%,100%,100%,100%)",
        plot_bgcolor="rgba(100%,100%,100%,100%)",
        xaxis_title=impact_title if impacts else "Pull",
        margin=dict(l=loffset, r=20, t=50, b=20),
        yaxis=dict(range=[-1, ndisplay]),
        showlegend=show_legend,
        legend=legend,
        legend_itemsizing="constant",
        height=100 * (ndisplay < 100)
        + ndisplay * 20.5
        + show_legend
        * (legend_pos != "right")
        * (impacts + pulls + asym_pulls)
        * (1 + include_ref)
        * 25,
        width=800 if show_legend and legend_pos == "right" else 640,
        font=dict(
            color="black",
        ),
    )

    gridargs = dict(
        showgrid=True,
        gridwidth=1,
        gridcolor="Gray",
        griddash="dash",
        zeroline=True,
        zerolinewidth=2,
        zerolinecolor="Gray",
    )
    tickargs = dict(
        tick0=0.0,
        tickmode="linear",
        tickangle=0,
        side="top",
    )

    text_on_bars = False
    labels = df["label"]
    if impacts and show_numbers:
        if include_ref:
            # append numerical values of impacts on nuisance name; fill up empty room with spaces to align numbers
            frmt = (
                "{:0"
                + str(
                    int(
                        max(0, np.log10(max(df["absimpact"])))
                        if max(df[f"absimpact_ref"]) > 0
                        else 0
                    )
                    + 2
                )
                + ".2f}"
            )
            nval = df["absimpact"].apply(
                lambda x, frmt=frmt: frmt.format(x)
            )  # .astype(str)
            nspace = nval.apply(
                lambda x, n=nval.apply(len).max(): " " * (n - len(x) + 1)
            )
            if include_ref:
                frmt_ref = (
                    "{:0"
                    + str(
                        int(
                            max(0, np.log10(max(df[f"absimpact_ref"])))
                            if max(df[f"absimpact_ref"]) > 0
                            else 0
                        )
                        + 2
                    )
                    + ".2f}"
                )
                nval_ref = df[f"absimpact_ref"].apply(
                    lambda x, frmt=frmt_ref: " (" + frmt.format(x) + ")"
                )
                nspace_ref = nval_ref.apply(
                    lambda x, n=nval_ref.apply(len).max(): " " * (n - len(x))
                )
                nval = nval + nspace_ref + nval_ref
            labels = labels + nspace + nval
        else:
            text_on_bars = True

    if impacts:

        def make_bar(
            key="impact",
            color="#377eb8",
            name="+1σ impact",
            text_on_bars=False,
            filled=True,
            opacity=1,
        ):
            x = np.where(df[key] < 0, np.nan, df[key]) if oneSidedImpacts else df[key]

            if text_on_bars:
                text = np.where(np.isnan(x), None, [f"{value:.2f}" for value in x])
            else:
                text = None

            return go.Bar(
                orientation="h",
                x=x,
                y=labels,
                text=text,
                textposition="outside",
                **get_marker(filled=filled, color=color, opacity=opacity),
                name=name,
            )

        sign = "+/-" if (oneSidedImpacts and not any(df["impact_down"] > 0)) else "+"
        label = f"{sign}1σ impact ({name})" if name else f"{sign}1σ impact"
        fig.add_trace(
            make_bar(
                key="impact_up",
                name=label,
                text_on_bars=text_on_bars,
                opacity=0.5 if include_ref else 1,
            ),
            row=1,
            col=1,
        )
        if include_ref:
            fig.add_trace(
                make_bar(
                    key="impact_up_ref",
                    name=f"{sign}1σ impact ({ref_name})",
                    filled=False,
                ),
                row=1,
                col=1,
            )

        if (oneSidedImpacts and any(df["impact_down"] > 0)) or not oneSidedImpacts:
            fig.add_trace(
                make_bar(
                    key="impact_down",
                    name=f"-1σ impact ({name})" if name else "-1σ impact",
                    color="#e41a1c",
                    text_on_bars=text_on_bars,
                    opacity=0.5 if include_ref else 1,
                ),
                row=1,
                col=1,
            )
            if include_ref:
                fig.add_trace(
                    make_bar(
                        key="impact_down_ref",
                        name=f"-1σ impact ({ref_name})",
                        color="#e41a1c",
                        filled=False,
                    ),
                    row=1,
                    col=1,
                )

        impact_range = df["absimpact"].max()
        if include_ref:
            impact_range = max(impact_range, df[f"absimpact_ref"].max())

        tick_spacing = math.pow(10, math.floor(math.log(impact_range, 10)))
        if tick_spacing > impact_range / 2:
            tick_spacing /= 2
        elif tick_spacing * 2 < impact_range:
            tick_spacing *= int(impact_range / (2 * tick_spacing))

        fig.update_layout(barmode="overlay")
        fig.update_layout(
            xaxis=dict(
                range=[
                    -impact_range * 1.2 if not oneSidedImpacts else -impact_range / 20,
                    impact_range * 1.2,
                ],
                dtick=tick_spacing,
                **gridargs,
                **tickargs,
            ),
        )

    if pulls:
        error_x = dict(
            color="black",
            thickness=1.5,
            width=5,
        )
        if asym:
            error_x["array"] = df["constraint_up"]
            error_x["arrayminus"] = df["constraint_down"]
        else:
            error_x["array"] = df["constraint"]

        fig.add_trace(
            go.Scatter(
                x=df["pull"],
                y=labels,
                mode="markers",
                marker=dict(
                    color="black",
                    size=8,
                ),
                error_x=error_x,
                name="Pulls ± Constraints",
                showlegend=include_ref,
            ),
            row=1,
            col=ncols,
        )
        if include_ref:
            if asym:
                base = df["pull_ref"] - df["constraint_down"]
                x = df["constraint_up"] + df["constraint_down"]
            else:
                base = df["pull_ref"] - df["constraint_ref"]
                x = 2 * df["constraint_ref"]

            fig.add_trace(
                go.Bar(
                    base=base,
                    x=x,
                    y=labels,
                    orientation="h",
                    **get_marker(filled=False, color="black"),
                    name=f"Pulls ± Constraints ({ref_name})",
                    showlegend=True,
                ),
                row=1,
                col=ncols,
            )

        if asym_pulls:
            fig.add_trace(
                go.Scatter(
                    x=df["newpull"],
                    y=labels,
                    mode="markers",
                    marker=dict(
                        color="green",
                        symbol="x",
                        size=8,
                        # line=dict(width=1),  # Adjust the thickness of the marker lines
                    ),
                    name="Asym. pulls",
                    showlegend=include_ref,
                ),
                row=1,
                col=ncols,
            )

            if include_ref:
                fig.add_trace(
                    go.Scatter(
                        x=df["newpull_ref"],
                        y=labels,
                        mode="markers",
                        marker=dict(
                            color="green",
                            symbol="circle-open",
                            size=8,
                            line=dict(
                                width=1
                            ),  # Adjust the thickness of the marker lines
                        ),
                        name=f"Asym. pulls ({ref_name})",
                        showlegend=include_ref,
                    ),
                    row=1,
                    col=ncols,
                )
        max_pull = np.max(df["abspull"])
        if pullrange is None:
            # Round up to nearest 0.5, add 1.1 for display
            pullrange = 0.5 * np.ceil(max_pull) + 1.1
        # Keep it a factor of 0.5, but no bigger than 1
        spacing = min(1, np.ceil(pullrange) / 2.0)
        if spacing > 0.5 * pullrange:  # make sure to have at least two ticks
            spacing /= 2.0
        xaxis_title = "Nuisance parameter"
        if diff_pulls:
            xaxis_title += "<br> θ-θ₀"
        else:
            xaxis_title += "<br> θ"
        info = dict(
            xaxis=dict(
                range=[-pullrange, pullrange], dtick=spacing, **gridargs, **tickargs
            ),
            xaxis_title=xaxis_title,
            yaxis=dict(range=[-1, ndisplay]),
            yaxis_visible=not impacts,
        )
        if impacts:
            new_info = {}
            for k in info.keys():
                new_info[k.replace("axis", "axis2")] = info[k]
            info = new_info
        fig.update_layout(barmode="overlay", **info)

    if title is not None:
        fig.add_annotation(
            x=0,
            y=1,
            xshift=-loffset,
            yshift=50,
            xref="paper",
            yref="paper",
            showarrow=False,
            text=title,
            font=dict(size=24, color="black", family="Arial", weight="bold"),
        )
        if subtitle is not None:
            fig.add_annotation(
                x=0,
                y=1,
                xshift=-loffset,
                yshift=25,
                xref="paper",
                yref="paper",
                showarrow=False,
                text=f"<i>{subtitle}</i>",
                font=dict(
                    size=20,
                    color="black",
                    family="Arial",
                ),
            )

    return fig


def readFitInfoFromFile(
    fitresult,
    poi,
    group=False,
    impact_type=False,
    grouping=None,
    asym=False,
    filters=[],
    stat=0.0,
    normalize=False,
    scale=1,
    diff_pulls=True,
):
    if poi is not None:
        out = io_tools.read_impacts_poi(
            fitresult,
            poi,
            group,
            pulls=not group,
            asym=asym,
            impact_type=impact_type,
            add_total=group and not impact_type == "nonprofiled",
        )
        if group:
            impacts, labels = out
            if normalize:
                idx = np.argwhere(labels == "Total")
                impacts /= impacts[idx].flatten()
        else:
            pulls, pulls_prefit, constraints, constraints_prefit, impacts, labels = out
            if normalize:
                imp, lab = io_tools.read_impacts_poi(
                    fitresult,
                    poi,
                    True,
                    pulls=False,
                    asym=asym,
                    impact_type=impact_type,
                    add_total=True,
                )

                idx = np.argwhere(lab == "Total")
                impacts /= imp[idx].flatten()

        if stat > 0 and "stat" in labels:
            idx = np.argwhere(labels == "stat")
            impacts[idx] = stat
    else:
        labels = io_tools.get_syst_labels(fitresult)
        _, pulls, constraints = io_tools.get_pulls_and_constraints(fitresult, asym=asym)
        _, pulls_prefit, constraints_prefit = io_tools.get_pulls_and_constraints(
            fitresult, prefit=True
        )

    apply_mask = (group and grouping is not None) or filters is not None

    if apply_mask:
        mask = np.ones(len(labels), dtype=bool)

        if group and grouping:
            mask &= np.isin(labels, grouping)  # Check if labels are in the grouping

        if filters:
            mask &= np.array(
                [any(re.search(f, label) for f in filters) for label in labels]
            )  # Apply regex filter

        labels = labels[mask]

    df = pd.DataFrame(np.array(labels, dtype=str), columns=["label"])

    if poi is not None:
        if apply_mask:
            impacts = impacts[mask]

        if scale and not normalize:
            impacts = impacts * scale

        if asym:
            df["impact_down"] = impacts[..., 1]
            df["impact_up"] = impacts[..., 0]
            df["absimpact"] = np.abs(impacts).max(axis=-1)
        else:
            df["impact_down"] = -impacts
            df["impact_up"] = impacts
            df["absimpact"] = np.abs(impacts)

    if not group:
        if apply_mask:
            pulls = pulls[mask]
            constraints = constraints[mask]
            pulls_prefit = pulls_prefit[mask]
            constraints_prefit = constraints_prefit[mask]

        df["pull_prefit"] = pulls_prefit

        if diff_pulls:
            df["pull"] = pulls - pulls_prefit
        else:
            df["pull"] = pulls
        df["abspull"] = np.abs(df["pull"])

        if asym:
            df["constraint_down"] = -constraints[..., 1]
            df["constraint_up"] = constraints[..., 0]
        else:
            df["constraint"] = constraints
            valid = (1 - constraints**2) > 0
            df["newpull"] = 999.0
            df.loc[valid, "newpull"] = df.loc[valid]["pull"] / np.sqrt(
                1 - df.loc[valid]["constraint"] ** 2
            )

    if poi:
        df = df.drop(df.loc[df["label"] == poi].index)

    return df


def readHistImpacts(
    fitresult,
    hist_impacts,
    hist_total,
    group=False,
    grouping=None,
    asym=False,
    filters=[],
    stat=0.0,
    normalize=False,
    scale=1,
):
    labels = np.array(hist_impacts.axes["impacts"])
    impacts = hist_impacts.values()

    if not group:
        labels_pulls, pulls, constraints = io_tools.get_pulls_and_constraints(
            fitresult, asym=asym
        )
        labels_pulls_prefit, pulls_prefit, constraints_prefit = (
            io_tools.get_pulls_and_constraints(fitresult, asym=asym, prefit=True)
        )

        labels, idxs, idxs_pulls = np.intersect1d(
            labels, labels_pulls, assume_unique=True, return_indices=True
        )

        pulls = pulls[idxs_pulls]
        constraints = constraints[idxs_pulls]
        pulls_prefit = pulls_prefit[idxs_pulls]
        constraints_prefit = constraints_prefit[idxs_pulls]

        impacts = impacts[idxs]

    total = np.sqrt(hist_total.variance)
    if group:
        impacts = np.append(impacts, total)
        labels = np.append(labels, "Total")

    # if args.relative:
    #     unit = "rel. unc. in %"
    #     impacts /= hist_total.value
    #     scale=100
    # else:
    #     unit = "bin unc."
    #     scale=1

    if normalize:
        impacts /= total

    if stat > 0 and "stat" in labels:
        idx = np.argwhere(labels == "stat")
        impacts[idx] = stat

    apply_mask = (group and grouping is not None) or filters is not None

    if apply_mask:
        mask = np.ones(len(labels), dtype=bool)

        if group and grouping:
            mask &= np.isin(labels, grouping)  # Check if labels are in the grouping

        if filters:
            mask &= np.array(
                [any(re.search(f, label) for f in filters) for label in labels]
            )  # Apply regex filter

        labels = labels[mask]

    df = pd.DataFrame(np.array(labels, dtype=str), columns=["label"])

    if apply_mask:
        impacts = impacts[mask]

    if scale and not normalize:
        impacts = impacts * scale

    if asym:
        df["impact_down"] = impacts[..., 1]
        df["impact_up"] = impacts[..., 0]
        df["absimpact"] = np.abs(impacts).max(axis=-1)
    else:
        df["impact_down"] = -impacts
        df["impact_up"] = impacts
        df["absimpact"] = np.abs(impacts)

    if not group:
        if apply_mask:
            pulls = pulls[mask]
            constraints = constraints[mask]
            pulls_prefit = pulls_prefit[mask]
            constraints_prefit = constraints_prefit[mask]

        df["pull"] = pulls
        df["pull_prefit"] = pulls_prefit
        df["pull"] = pulls - pulls_prefit
        df["abspull"] = np.abs(df["pull"])

        if asym:
            df["constraint_down"] = -constraints[..., 1]
            df["constraint_up"] = constraints[..., 0]
        else:
            df["constraint"] = constraints
            valid = (1 - constraints**2) > 0
            df["newpull"] = 999.0
            df.loc[valid, "newpull"] = df.loc[valid]["pull"] / np.sqrt(
                1 - df.loc[valid]["constraint"] ** 2
            )

    return df


def parseArgs():
    sort_choices = ["label", "pull", "abspull", "constraint", "absimpact"]
    sort_choices += [
        *[
            f"{c}_diff" for c in sort_choices
        ],  # possibility to sort based on largest difference between input and referencefile
        *[
            f"{c}_ref" for c in sort_choices
        ],  # possibility to sort based on reference file
        *[f"{c}_both" for c in sort_choices],
    ]  # possibility to sort based on the largest/smallest of both input and reference file

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "inputFile",
        type=str,
        help="fitresults output hdf5 file from fit",
    )
    parser.add_argument(
        "--result",
        default=None,
        type=str,
        help="fitresults key in file (e.g. 'asimov'). Leave empty for data fit result.",
    )
    parser.add_argument(
        "-m",
        "--physicsModel",
        default=None,
        type=str,
        nargs="+",
        help="Print impacts on observables use '-m <model> channel axes' for physics model results.",
    )
    parser.add_argument(
        "-r",
        "--referenceFile",
        type=str,
        help="fitresults output hdf5 file from fit for reference",
    )
    parser.add_argument(
        "--refResult",
        default=None,
        type=str,
        help="fitresults key in file (e.g. 'asimov'). Leave empty for data fit result.",
    )
    parser.add_argument(
        "--refName",
        type=str,
        help="Name of reference input for legend",
    )
    parser.add_argument(
        "--name",
        type=str,
        help="Name of input for legend",
    )
    parser.add_argument(
        "-s",
        "--sort",
        default="absimpact",
        type=str,
        help="Sort mode for nuisances",
        choices=sort_choices,
    )
    parser.add_argument(
        "--stat",
        default=0.0,
        type=float,
        help="Overwrite stat. uncertainty with this value",
    )
    parser.add_argument(
        "-d",
        "--sortDescending",
        dest="ascending",
        action="store_false",
        help="Sort mode for nuisances",
    )
    parser.add_argument(
        "--mode",
        choices=["group", "ungrouped", "both"],
        default="both",
        help="Impact mode",
    )
    parser.add_argument(
        "--normalize",
        action="store_true",
        help="Normalize impacts on poi, leading to relative uncertainties.",
    )
    parser.add_argument(
        "--relative",
        action="store_true",
        help="Relative uncertainty w.r.t. central value of parameter of interest (only for hisograms)",
    )
    parser.add_argument("--debug", action="store_true", help="Print debug output")
    parser.add_argument(
        "--diffPullAsym",
        action="store_true",
        help="Also add the pulls after the diffPullAsym definition",
    )
    parser.add_argument(
        "--oneSidedImpacts", action="store_true", help="Make impacts one-sided"
    )
    parser.add_argument(
        "--filters",
        nargs="*",
        type=str,
        help="Filter regexes to select nuisances by name",
    )
    parser.add_argument(
        "--grouping",
        type=str,
        default=None,
        help="Pre-defined grouping in config to select nuisance groups",
    )
    parser.add_argument(
        "--config",
        type=str,
        default=None,
        help="Path to config file for style formatting",
    )
    parser.add_argument(
        "--title",
        default="Rabbit",
        type=str,
        help="Title to be printed in upper left",
    )
    parser.add_argument(
        "--subtitle",
        default="",
        type=str,
        help="Subtitle to be printed after title",
    )
    parser.add_argument(
        "--impactTitle",
        default="Impacts",
        type=str,
        help="Title for impacts",
    )
    parser.add_argument("--noImpacts", action="store_true", help="Don't show impacts")
    parser.add_argument(
        "--globalImpacts", action="store_true", help="Print global impacts"
    )
    parser.add_argument(
        "--nonprofiledImpacts", action="store_true", help="Print non-profiled impacts"
    )
    parser.add_argument(
        "--asymImpacts",
        action="store_true",
        help="Print asymmetric impacts from likelihood, otherwise symmetric from hessian",
    )
    parser.add_argument(
        "--showNumbers", action="store_true", help="Show values of impacts"
    )
    parser.add_argument(
        "--poi",
        type=str,
        default=None,
        help="Specify POI to make impacts for, otherwise use all",
    )
    parser.add_argument(
        "--poiType", type=str, default=None, help="POI type to make impacts for"
    )
    parser.add_argument(
        "--pullrange", type=float, default=None, help="POI type to make impacts for"
    )
    parser.add_argument(
        "-o",
        "--outpath",
        type=str,
        default="./test",
        help="Folder path for output",
    )
    parser.add_argument(
        "-p", "--postfix", type=str, help="Postfix for output file name"
    )
    parser.add_argument(
        "--eoscp",
        action="store_true",
        help="Override use of xrdcp and use the mount instead",
    )
    parser.add_argument(
        "--otherExtensions",
        default=[],
        type=str,
        nargs="*",
        help="Additional output file types to write",
    )
    parser.add_argument(
        "-n", "--num", type=int, default=50, help="Number of nuisances to plot"
    )
    parser.add_argument(
        "--noPulls",
        action="store_true",
        help="Don't show pulls (not defined for groups)",
    )
    parser.add_argument(
        "--scaleImpacts",
        type=float,
        default=1.0,
        help="Scale impacts by this number",
    )
    parser.add_argument(
        "--pullsNoDiff",
        action="store_true",
        help="Plot actual nuisance parameter value, by default nuisance parameter difference w.r.t. prefit value",
    )
    return parser.parse_args()


def make_plots(
    df,
    args,
    outdir,
    outfile,
    group=False,
    asym=False,
    pullrange=None,
    meta=None,
    postfix=None,
    impact_title=None,
):

    df = df.fillna(0)

    if args.sort:
        if args.sort.endswith("diff"):
            key = args.sort.replace("_diff", "")
            df[f"{key}_diff"] = abs(df[key] - df[f"{key}_ref"])
        elif args.sort.endswith("both"):
            key = args.sort.replace("_both", "")
            if args.ascending:
                df[f"{key}_both"] = df[[key, f"{key}_ref"]].max(axis=1)
            else:
                df[f"{key}_both"] = df[[key, f"{key}_ref"]].min(axis=1)

        if args.sort in df.keys():
            if f"{args.sort}_ref" in df.keys():
                df = df.sort_values(
                    by=[args.sort, f"{args.sort}_ref"], ascending=args.ascending
                )
            else:
                df = df.sort_values(by=args.sort, ascending=args.ascending)
        else:
            print(
                f"Trying to sort {args.sort} but not found in dataframe, continue without sorting"
            )

    outfile = os.path.join(outdir, outfile)
    extensions = [outfile.split(".")[-1], *args.otherExtensions]

    include_ref = any(
        x in df.keys()
        for x in ["impact_ref", "absimpact_ref", "pull_ref", "constraint_ref"]
    )

    kwargs = dict(
        pulls=not args.noPulls and not group,
        impact_title=impact_title,
        oneSidedImpacts=args.oneSidedImpacts,
        pullrange=pullrange,
        title=args.title,
        subtitle=args.subtitle,
        impacts=not args.noImpacts,
        asym=asym,
        asym_pulls=args.diffPullAsym,
        include_ref=include_ref,
        ref_name=args.refName,
        name=args.name,
        show_numbers=args.showNumbers,
        show_legend=(not group and not args.noImpacts)
        or include_ref
        or args.name is not None,
        group=group,
        diff_pulls=not args.pullsNoDiff,
    )

    if args.num and args.num < int(df.shape[0]):
        # in case multiple extensions are given including html, don't do the skimming on html but all other formats
        if "html" in extensions and len(extensions) > 1:
            fig = plotImpacts(df, legend_pos="right", **kwargs)
            outfile_html = ".".join([*outfile.split(".")[:-1], "html"])
            writeOutput(fig, outfile_html, [".html"], postfix=postfix)
            extensions = [e for e in extensions if e != "html"]
            outfile = ".".join([*outfile.split(".")[:-1], extensions[0]])

        df = df[-args.num :]

    fig = plotImpacts(df, **kwargs)

    writeOutput(fig, outfile, extensions, postfix=postfix, args=args, meta_info=meta)


def load_dataframe_parms(
    args,
    fitresult,
    poi=None,
    group=False,
    asym=False,
    normalize=False,
    fitresult_ref=None,
    grouping=None,
    translate_label={},
):
    if args.globalImpacts:
        impact_type = "global"
    elif args.nonprofiledImpacts:
        impact_type = "nonprofiled"
    else:
        impact_type = "traditional"

    if not group:
        df = readFitInfoFromFile(
            fitresult,
            poi,
            False,
            asym=asym,
            impact_type=impact_type,
            filters=args.filters,
            stat=args.stat / 100.0,
            normalize=normalize,
            scale=args.scaleImpacts,
            diff_pulls=not args.pullsNoDiff,
        )
    elif group:
        df = readFitInfoFromFile(
            fitresult,
            poi,
            True,
            asym=asym,
            impact_type=impact_type,
            filters=args.filters,
            stat=args.stat / 100.0,
            normalize=normalize,
            grouping=grouping,
            scale=args.scaleImpacts,
        )

    if fitresult_ref:
        df_ref = readFitInfoFromFile(
            fitresult_ref,
            poi,
            group,
            asym=asym,
            impact_type=impact_type,
            filters=args.filters,
            stat=args.stat / 100.0,
            normalize=normalize,
            grouping=grouping,
            scale=args.scaleImpacts,
            diff_pulls=not args.pullsNoDiff,
        )
        df = df.merge(df_ref, how="outer", on="label", suffixes=("", "_ref"))

    df["label"] = df["label"].apply(lambda l: translate_label.get(l, l))

    if df.empty:
        print("WARNING: Empty dataframe")
        if group and grouping:
            print(
                f"WARNING: This can happen if no group is found that belongs to {grouping}"
            )
            print(
                "WARNING: Try a different mode for --grouping or use '--mode ungrouped' to skip making impacts for groups"
            )
        print("WARNING: Skipping this part")
        return None
    else:
        return df


def load_dataframe_hists(
    fitresult,
    args,
    hist_impacts,
    hist_total,
    ibin=None,
    group=False,
    asym=False,
    normalize=False,
    fitresult_ref=None,
    hist_impacts_ref=None,
    hist_total_ref=None,
    grouping=None,
    translate_label={},
):

    if args.relative:
        scale = args.scaleImpacts * 1.0 / hist_total.value
        if fitresult_ref:
            scale_ref = args.scaleImpacts * 1.0 / hist_total_ref.value
    else:
        scale = args.scaleImpacts
        if fitresult_ref:
            scale_ref = args.scaleImpacts

    df = readHistImpacts(
        fitresult,
        hist_impacts,
        hist_total,
        group,
        filters=args.filters,
        stat=args.stat / 100.0,
        normalize=normalize,
        grouping=grouping,
        scale=scale,
    )

    if fitresult_ref:
        df_ref = readHistImpacts(
            fitresult_ref,
            hist_impacts_ref,
            hist_total_ref,
            group,
            filters=args.filters,
            stat=args.stat / 100.0,
            normalize=normalize,
            grouping=grouping,
            scale=scale_ref,
        )
        df = df.merge(df_ref, how="outer", on="label", suffixes=("", "_ref"))

    df["label"] = df["label"].apply(lambda l: translate_label.get(l, l))

    if df.empty:
        print("WARNING: Empty dataframe")
        if group and grouping:
            print(
                f"WARNING: This can happen if no group is found that belongs to {grouping}"
            )
            print(
                "WARNING: Try a different mode for --grouping or use '--mode ungrouped' to skip making impacts for groups"
            )
        print("WARNING: Skipping this part")
        return None
    else:
        return df


def produce_plots_parms(
    args,
    fitresult,
    outdir,
    outfile,
    poi=None,
    group=False,
    asym=False,
    normalize=False,
    fitresult_ref=None,
    pullrange=None,
    meta=None,
    postfix=None,
    impact_title=None,
    grouping=None,
    translate_label={},
):

    df = load_dataframe_parms(
        args,
        fitresult,
        poi=poi,
        group=group,
        asym=asym,
        normalize=normalize,
        fitresult_ref=fitresult_ref,
        grouping=grouping,
        translate_label=translate_label,
    )

    if df is None:
        return

    make_plots(
        df,
        args,
        outdir,
        outfile,
        group=group,
        asym=asym,
        pullrange=pullrange,
        meta=meta,
        postfix=postfix,
        impact_title=impact_title,
    )


def produce_plots_hist(
    args,
    fitresult,
    outdir,
    outfile,
    hist_impacts,
    hist_total,
    ibin=None,
    group=False,
    asym=False,
    normalize=False,
    fitresult_ref=None,
    hist_impacts_ref=None,
    hist_total_ref=None,
    pullrange=None,
    meta=None,
    postfix=None,
    impact_title=None,
    grouping=None,
    translate_label={},
):

    df = load_dataframe_hists(
        fitresult,
        args,
        hist_impacts,
        hist_total,
        ibin=ibin,
        group=group,
        asym=asym,
        normalize=normalize,
        fitresult_ref=fitresult_ref,
        hist_impacts_ref=hist_impacts_ref,
        hist_total_ref=hist_total_ref,
        grouping=grouping,
        translate_label=translate_label,
    )
    if df is None:
        return

    make_plots(
        df,
        args,
        outdir,
        outfile,
        group=group,
        asym=asym,
        pullrange=pullrange,
        meta=meta,
        postfix=postfix,
        impact_title=impact_title,
    )


def main():
    args = parseArgs()

    config = plot_tools.load_config(args.config)

    translate_label = getattr(config, "impact_labels", {})

    fitresult, meta = io_tools.get_fitresult(args.inputFile, args.result, meta=True)
    if args.referenceFile is not None or args.refResult is not None:
        referenceFile = (
            args.referenceFile if args.referenceFile is not None else args.inputFile
        )
        fitresult_ref = io_tools.get_fitresult(referenceFile, args.refResult)
    else:
        fitresult_ref = None

    meta_out = {
        "rabbit": meta["meta_info"],
    }

    outdir = output_tools.make_plot_dir(args.outpath, eoscp=args.eoscp)

    kwargs = dict(
        pullrange=args.pullrange,
        asym=args.asymImpacts,
        fitresult_ref=fitresult_ref,
        meta=meta_out,
        postfix=args.postfix,
        translate_label=translate_label,
    )

    if args.noImpacts:
        # do one pulls plot, ungrouped
        produce_plots_parms(args, fitresult, outdir, outfile="pulls.html", **kwargs)
    else:
        kwargs.update(dict(normalize=args.normalize, impact_title=args.impactTitle))

        impacts_name = "impacts"
        if args.globalImpacts:
            impacts_name = f"global_{impacts_name}"
        elif args.nonprofiledImpacts:
            impacts_name = f"nonprofiled_{impacts_name}"

        grouping = None
        if args.grouping is not None:
            grouping = getattr(config, "nuisance_grouping", {}).get(args.grouping, None)

        if args.physicsModel is not None:
            if args.asymImpacts:
                raise NotImplementedError(
                    "Asymetric impacts on observables is not yet implemented"
                )
            if not args.globalImpacts:
                raise NotImplementedError(
                    "Only global impacts on observables is implemented (use --globalImpacts)"
                )

            model_key = " ".join(args.physicsModel)
            if model_key in fitresult["physics_models"].keys():
                channels = fitresult["physics_models"][model_key]["channels"]
            else:
                keys = [
                    key
                    for key in fitresult["physics_models"].keys()
                    if key.startswith(model_key)
                ]
                channels = fitresult["physics_models"][keys[0]]["channels"]

            for channel, hists in channels.items():

                modes = ["ungrouped", "group"] if args.mode == "both" else [args.mode]
                for mode in modes:
                    group = mode == "group"

                    key = "hist_postfit_inclusive_global_impacts"
                    if group:
                        key += "_grouped"

                    hist_total = hists["hist_postfit_inclusive"].get()

                    hist = hists[key].get()

                    # TODO: implement ref
                    # hist_ref
                    # hist_total_ref

                    if fitresult_ref is not None:
                        hists_ref = fitresult_ref["physics_models"][model_key][
                            "channels"
                        ][channel]
                        hist_ref = hists_ref[key].get()
                        hist_total_ref = hists_ref["hist_postfit_inclusive"].get()

                    for idxs in itertools.product(
                        *[np.arange(a.size) for a in hist_total.axes]
                    ):
                        ibin = {a: i for a, i in zip(hist_total.axes.name, idxs)}
                        print(f"Now at {ibin}")

                        ibin_str = "_".join([f"{a}{i}" for a, i in ibin.items()])
                        if group:
                            outfile = f"{impacts_name}_grouped_{ibin_str}.html"
                        else:
                            outfile = f"{impacts_name}_{ibin_str}.html"
                            if not args.noPulls:
                                outfile = f"pulls_and_{outfile}"

                        produce_plots_hist(
                            args,
                            fitresult,
                            outdir,
                            outfile,
                            hist[ibin],
                            hist_total[ibin],
                            ibin,
                            group=group,
                            grouping=grouping,
                            hist_impacts_ref=hist_ref[ibin] if fitresult_ref else None,
                            hist_total_ref=(
                                hist_total_ref[ibin] if fitresult_ref else None
                            ),
                            **kwargs,
                        )
        else:
            pois = [args.poi] if args.poi else io_tools.get_poi_names(meta)
            for poi in pois:
                print(f"Now at {poi}")
                if args.mode in ["both", "ungrouped"]:
                    name = f"{impacts_name}_{poi}.html"
                    if not args.noPulls:
                        name = f"pulls_and_{name}"
                    produce_plots_parms(
                        args, fitresult, outdir, outfile=name, poi=poi, **kwargs
                    )
                if args.mode in ["both", "group"]:
                    produce_plots_parms(
                        args,
                        fitresult,
                        outdir,
                        outfile=f"{impacts_name}_grouped_{poi}.html",
                        poi=poi,
                        group=True,
                        grouping=grouping,
                        **kwargs,
                    )

    output_tools.write_indexfile(outdir)

    if output_tools.is_eosuser_path(args.outpath) and args.eoscp:
        output_tools.copy_to_eos(outdir, args.outpath)


if __name__ == "__main__":
    main()



import os

import hist
import numpy as np
from tqdm import tqdm
from wums import logging

from rabbit import tensorwriter
from rabbit.datacard_parser import DatacardParser

logger = logging.child_logger(__name__)


class DatacardConverter:
    """
    Convert data from Combine datacards and ROOT files to hdf5 tensor format
    """

    def __init__(self, datacard_file, symmetrize=None, use_root=False, mass="125.38"):
        """
        Initialize the converter with a datacard file

        Args:
            datacard_file: Path to the datacard file
        """
        self.datacard_file = datacard_file
        self.parser = DatacardParser()
        self.mass = mass
        self.symmetrize = symmetrize

        # For counting experiments
        self.yield_axis = hist.axis.Integer(
            0, 1, name="yield", overflow=False, underflow=False
        )

        self.use_root = use_root
        if self.use_root:
            import ROOT
            from narf.histutils import root_to_hist

            self.io = ROOT
            self.root_to_hist = root_to_hist
            self.root_files = (
                {}
            )  # Cache for opened ROOT files and their directories as uproot objects
        else:
            import uproot

            self.io = uproot
            self.root_directories = {}  # Cache for opened ROOT files

    def parse(self):
        """Parse the datacard file"""
        self.parser.parse_file(self.datacard_file)
        return self

    def get_root_file(self, file_path):
        """Get a ROOT file, opening it if not already open"""
        if file_path not in self.root_files:
            # Check if path is relative to the datacard location
            datacard_dir = os.path.dirname(os.path.abspath(self.datacard_file))
            if not os.path.isabs(file_path):
                file_path = os.path.join(datacard_dir, file_path)

            if not os.path.exists(file_path):
                raise FileNotFoundError(f"ROOT file not found: {file_path}")

            self.root_files[file_path] = self.io.TFile.Open(file_path, "READ")

        return self.root_files[file_path]

    def load_root_directories(self, file_path):
        """Get a ROOT file, opening it if not already open"""
        if file_path not in self.root_directories:
            # Check if path is relative to the datacard location
            datacard_dir = os.path.dirname(os.path.abspath(self.datacard_file))
            if not os.path.isabs(file_path):
                file_path = os.path.join(datacard_dir, file_path)

            if not os.path.exists(file_path):
                raise FileNotFoundError(f"ROOT file not found: {file_path}")

            logger.info(f"Open file {file_path}")

            file = self.io.open(file_path)

            search_depth = self.parser.max_depth - 1
            if search_depth >= 0:
                # recursively find all directories up to a depth of 'search_depth', setting 'search_depth' to a proper value speeds up things (if you know how deep the directories go)
                def get_directories(directory, level=0):
                    keys = set([k.split("/")[0] for k in directory.keys()])
                    directories = [
                        d
                        for d in keys
                        if isinstance(directory[d], self.io.reading.ReadOnlyDirectory)
                    ]
                    if len(directories) == 0:
                        return directory
                    elif level >= search_depth:
                        return {d: directory[d] for d in directories}
                    else:
                        return {
                            d: get_directories(directory[d], level=level + 1)
                            for d in directories
                        }

                self.root_directories[file_path] = get_directories(file)
            else:
                self.root_directories[file_path] = file

    def is_TH1(self, histogram):
        if self.use_root:
            return isinstance(histogram, self.io.TH1)
        else:
            return histogram.classname.startswith("TH1")

    def get_histogram(
        self,
        shape_info,
        process,
        bin_name,
        systematic=None,
        demand=True,
        edges=False,
        variances=None,
    ):
        """
        Get a histogram based on shape information

        Args:
            shape_info: Shape directive information
            process: Process name
            bin_name: Bin/channel name
            systematic: Systematic uncertainty (None for nominal)
            demand: throw an error if shape is not found and demand=True, otherwise not

        Returns:
            TH1 histogram or None if not found
        """
        if not shape_info:
            return None

        file_path = shape_info["file"]

        if systematic and "histogram_syst_pattern" in shape_info:
            syst_pattern = shape_info["histogram_syst_pattern"]
            hist_name = syst_pattern.replace("$PROCESS", process)
            hist_name = hist_name.replace("$CHANNEL", bin_name)
            hist_name = hist_name.replace("$SYSTEMATIC", systematic)
            hist_name = hist_name.replace("$MASS", self.mass)
        else:
            # Replace variables in histogram pattern
            hist_pattern = shape_info["histogram_pattern"]
            hist_name = hist_pattern.replace("$PROCESS", process)
            hist_name = hist_name.replace("$CHANNEL", bin_name)
            hist_name = hist_name.replace("$MASS", self.mass)

            if systematic:
                # If no separate systematic pattern, assume appending Up/Down to nominal
                hist_name = f"{hist_name}_{systematic}"

        if self.use_root:
            root_file = self.get_root_file(file_path)
            histogram = root_file.Get(hist_name)
        else:
            # Try to find the histogram in the file
            histogram = self.root_directories[file_path]
            for p in hist_name.split("/"):
                histogram = histogram.get(p, {})

        if not histogram or not self.is_TH1(histogram):
            if demand:
                raise ValueError(f"Histogram {hist_name} not found in {file_path}")
            else:
                logger.debug(
                    f"Histogram {hist_name} not found in {file_path}, but 'demand=False'. Skip it"
                )
                return None
        else:
            if self.use_root:
                h = self.root_to_hist(histogram)
                histogram.Delete()
                return h
            else:
                return histogram.to_hist()

    def convert_to_hdf5(self, sparse=False):
        """
        Convert the datacard and histograms to numpy arrays

        Returns:
            Dictionary of numpy arrays with all the data
        """
        logger.info("Parse datacard text")
        self.parse()
        logger.info("Prepare histograms")
        # Get shape directives for each process and bin
        shape_map = {}
        for shape in self.parser.shapes:
            process = shape["process"]
            channel = shape["channel"]

            # Handle wildcards
            if process == "*":
                processes = ["data_obs", *self.parser.processes]
            else:
                processes = [process]

            if channel == "*":
                channels = self.parser.bins
            else:
                channels = [channel]

            for p in processes:
                for c in channels:
                    shape_map[(p, c)] = shape

            if not self.use_root:
                self.load_root_directories(shape["file"])

        logger.info("Convert histograms into hdf5 tensor")

        # TODO: nuisance groups

        writer = tensorwriter.TensorWriter(
            sparse=sparse,
        )

        logger.info("loop over channels (aka combine bins) for nominal histograms")
        for bin_name in tqdm(self.parser.bins, desc="Processing"):

            for process_name in ["data_obs", *self.parser.bin_process_map[bin_name]]:

                shape_info = (
                    shape_map.get((process_name, bin_name))
                    or shape_map.get(("*", bin_name))
                    or shape_map.get((process_name, "*"))
                    or shape_map.get(("*", "*"))
                )

                if shape_info:
                    h_proc = self.get_histogram(
                        shape_info,
                        process_name,
                        bin_name,
                    )
                else:
                    # For counting experiments
                    h_proc = hist.Hist(
                        self.yield_axis,
                        data=np.array([self.parser.observations.get(bin_name, 0)]),
                        storage=hist.storage.Double(),
                    )

                if process_name == "data_obs":
                    writer.add_channel(h_proc.axes, bin_name)
                    writer.add_data(h_proc, bin_name)
                else:
                    writer.add_process(
                        h_proc,
                        process_name,
                        bin_name,
                        signal=self.parser.process_indices.get(process_name, 0) <= 0,
                    )

        # Add rate parameters as unconstrained lnN uncertainties with 1% variation.
        #   To compare with combine use 'r = exp(100 * r)'
        for parts in self.parser.rate_params:
            name = parts[0]
            channels = [parts[1]] if parts[1] != "*" else self.parser.bins
            processes = [parts[2]] if parts[2] != "*" else self.parser.processes
            # TODO: set initial value
            value = float(parts[3])
            # TODO: use parameter range
            if len(parts) >= 5:
                lo, hi = parts[4][1:-1].split(",")
                prange = (float(lo), float(hi))

            for c in channels:
                writer.add_norm_systematic(name, processes, c, 1.01, constrained=False)

        def add_norm_syst(writer, name, process, channel, effect):
            # Parse the effect (could be asymmetric like 0.9/1.1)
            if "/" in effect:
                down, up = effect.split("/")
                writer.add_norm_systematic(
                    name, process, channel, [(float(up), float(down))]
                )
            else:
                writer.add_norm_systematic(name, process, channel, float(effect))

        logger.info("loop over systematic variations")
        for syst in tqdm(self.parser.systematics, desc="Processing"):

            if syst["type"] in ["shape", "shapeN", "shape?", "shapeU"]:
                # TODO dedicated treatment for shapeN
                for (bin_name, process_name), effect in syst["effects"].items():
                    if effect not in ["-", "0"]:

                        shape_info = (
                            shape_map.get((process_name, bin_name))
                            or shape_map.get(("*", bin_name))
                            or shape_map.get((process_name, "*"))
                            or shape_map.get(("*", "*"))
                        )

                        hist_up = self.get_histogram(
                            shape_info,
                            process_name,
                            bin_name,
                            f"{syst['name']}Up",
                            demand=syst["type"] != "shape?",
                        )
                        hist_down = self.get_histogram(
                            shape_info,
                            process_name,
                            bin_name,
                            f"{syst['name']}Down",
                            demand=syst["type"] != "shape?",
                        )

                        if hist_up is None and hist_down is None:
                            # 'syst?' case
                            add_norm_syst(
                                writer, syst["name"], process_name, bin_name, effect
                            )
                        else:
                            writer.add_systematic(
                                [hist_up, hist_down],
                                syst["name"],
                                process_name,
                                bin_name,
                                kfactor=float(effect),
                                constrained=syst["type"]
                                != "shapeU",  # TODO check if shapeU is unconstriained
                                symmetrize=self.symmetrize,
                            )

            elif syst["type"] in ["lnN", "lnU"]:
                # TODO dedicated treatment for lnU

                for (bin_name, process_name), effect in syst["effects"].items():
                    if effect not in ["-", "0"]:
                        add_norm_syst(
                            writer, syst["name"], process_name, bin_name, effect
                        )

        return writer

    def __del__(self):
        if self.use_root:
            logger.info("Close root files")
            for file in self.root_files.values():
                file.Close()



#!/usr/bin/env python3

import argparse
import os

import matplotlib.pyplot as plt
import mplhep as hep
import numpy as np

from rabbit import io_tools

from wums import logging, output_tools, plot_tools  # isort: skip

hep.style.use(hep.style.ROOT)

logger = None


def writeOutput(fig, outfile, extensions=[], postfix=None, args=None, meta_info=None):
    name, _ = os.path.splitext(outfile)

    if postfix:
        name += f"_{postfix}"

    for ext in extensions:
        if ext[0] != ".":
            ext = "." + ext
        output = name + ext
        print(f"Write output file {output}")
        plt.savefig(output)

        output = name.rsplit("/", 1)
        output[1] = os.path.splitext(output[1])[0]
        if len(output) == 1:
            output = (None, *output)
    if args is None and meta_info is None:
        return
    output_tools.write_logfile(
        *output,
        args=args,
        meta_info=meta_info,
    )


def parseArgs():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "inputFile",
        type=str,
        help="fitresults output",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        type=int,
        default=3,
        choices=[0, 1, 2, 3, 4],
        help="Set verbosity level with logging, the larger the more verbose",
    )
    parser.add_argument(
        "--noColorLogger", action="store_true", help="Do not use logging with colors"
    )
    parser.add_argument(
        "--result",
        default=None,
        type=str,
        help="fitresults key in file (e.g. 'asimov'). Leave empty for data fit result.",
    )
    parser.add_argument(
        "-o",
        "--outpath",
        type=str,
        default="./test",
        help="Folder path for output",
    )
    parser.add_argument(
        "-p", "--postfix", type=str, help="Postfix for output file name"
    )
    parser.add_argument(
        "--params",
        type=str,
        nargs="*",
        default=[],
        help="Parameters to plot the likelihood scan",
    )
    parser.add_argument(
        "--combine",
        type=str,
        default=None,
        help="Provide a root file with likelihood scan result from combine",
    )
    parser.add_argument(
        "--title",
        default="Rabbit",
        type=str,
        help="Title to be printed in upper left",
    )
    parser.add_argument(
        "--subtitle",
        default="",
        type=str,
        help="Subtitle to be printed after title",
    )
    parser.add_argument(
        "--xlim",
        type=float,
        nargs=2,
        default=None,
        help="x axis limits",
    )
    parser.add_argument("--titlePos", type=int, default=2, help="title position")
    parser.add_argument(
        "--config",
        type=str,
        default=None,
        help="Path to config file for style formatting",
    )
    return parser.parse_args()


def plot_scan(
    h_scan,
    h_contours=None,
    param_value=0,
    param_variance=1,
    param="x",
    title=None,
    subtitle=None,
    titlePos=0,
    xlim=None,
    ylabel=r"$-2\,\Delta \log L$",
    config={},
    combine=None,
):

    xlabel = getattr(config, "systematics_labels", {}).get(param, param)

    mask = np.isfinite(h_scan.values())

    x = np.array(h_scan.axes["scan"]).astype(float)[mask]
    y = h_scan.values()[mask] * 2

    fig, ax = plot_tools.figure(
        x,
        xlabel,
        ylabel,
        xlim=(min(x), max(x)),
        ylim=(min(y), max(y)),  # logy=args.logy
    )

    ax.axhline(y=1, color="gray", linestyle="--", alpha=0.5)
    ax.axhline(y=4, color="gray", linestyle="--", alpha=0.5)

    parabola_vals = param_value + np.linspace(
        -3 * param_variance**0.5, 3 * param_variance**0.5, 100
    )
    parabola_nlls = 1 / param_variance * (parabola_vals - param_value) ** 2
    ax.plot(
        parabola_vals,
        parabola_nlls,
        marker="",
        markerfacecolor="none",
        color="red",
        linestyle="-",
        label="Hessian",
    )

    ax.plot(
        x,
        y,
        marker="x",
        color="blue",
        label="Likelihood scan",
        markeredgewidth=2,
    )

    if combine is not None:
        ax.plot(*combine, marker="o", color="orange", label="Combine")

    if h_contours is not None:
        for i, cl in enumerate(h_contours.axes["confidence_level"]):
            x = h_contours[{"confidence_level": cl}].values()[::-1] + param_value
            y = np.full(len(x), float(cl) ** 2)
            label = "Contour scan" if i == 0 else None
            ax.plot(
                x,
                y,
                marker="o",
                markerfacecolor="none",
                color="black",
                linestyle="",
                markeredgewidth=2,
                label=label,
            )
            logger.info(f"{int(float(cl))} sigma confidence level for {param} in {x}")

            ax.axvline(x=0, color="gray", linestyle="--", alpha=0.5)
            for ix in x:
                ax.axvline(x=ix, color="gray", linestyle="--", alpha=0.5)

    ax.legend(loc="upper right")

    plot_tools.add_decor(
        ax,
        title,
        subtitle,
        data=False,
        lumi=None,
        loc=titlePos,
    )

    return fig


def main():
    args = parseArgs()

    global logger
    logger = logging.setup_logger(__file__, args.verbose, args.noColorLogger)

    fitresult, meta = io_tools.get_fitresult(args.inputFile, args.result, meta=True)

    config = plot_tools.load_config(args.config)

    meta = {
        "rabbit": meta["meta_info"],
    }

    h_params = fitresult["parms"].get()

    if "contour_scan" in fitresult.keys():
        h_contour = fitresult["contour_scan"].get()
    else:
        h_contour = None

    parms = h_params.axes["parms"] if len(args.params) == 0 else args.params

    if args.combine is not None:
        import uproot

        with uproot.open(args.combine) as tfile:
            vals = tfile["limit"]["r"].array()
            nlls = tfile["limit"]["deltaNLL"].array()

            order = np.argsort(vals)
            vals = vals[order]
            nlls = nlls[order] * 2  # -2ln(L)

    for param in parms:
        p = h_params[{"parms": param}]
        param_value = p.value
        param_variance = p.variance
        h_scan = fitresult[f"nll_scan_{param}"].get()

        if h_contour is not None:
            h_contour_param = h_contour[{"parms": param, "impacts": param}]
        else:
            h_contour_param = None

        fig = plot_scan(
            h_scan,
            h_contour_param,
            param_value=param_value,
            param_variance=param_variance,
            param=param,
            title=args.title,
            subtitle=args.subtitle,
            titlePos=args.titlePos,
            xlim=args.xlim,
            config=config,
            combine=(vals, nlls) if args.combine is not None else None,
        )
        os.makedirs(args.outpath, exist_ok=True)
        outfile = os.path.join(args.outpath, f"nll_scan_{param}")
        writeOutput(
            fig,
            outfile=outfile,
            extensions=["png", "pdf"],
            meta_info=meta,
            args=args,
            postfix=args.postfix,
        )


if __name__ == "__main__":
    main()



import argparse

import numpy as np

from rabbit import tensorwriter

parser = argparse.ArgumentParser()
parser.add_argument("-o", "--output", default="./", help="output directory")
parser.add_argument("--outname", default="test_tensor", help="output file name")
parser.add_argument(
    "--postfix",
    default=None,
    type=str,
    help="Postfix to append on output file name",
)
parser.add_argument(
    "--sparse",
    default=False,
    action="store_true",
    help="Make sparse tensor",
)
parser.add_argument(
    "--symmetrizeAll",
    default=False,
    action="store_true",
    help="Make fully symmetric tensor",
)
parser.add_argument(
    "--skipMaskedChannels",
    default=False,
    action="store_true",
    help="Skip adding masked channels",
)
parser.add_argument(
    "--systematicType",
    choices=["log_normal", "normal"],
    default="log_normal",
    help="probability density for systematic variations",
)
parser.add_argument(
    "--addSystToDataCovariance",
    default=False,
    action="store_true",
    help="Add systematics to data covariance matrix, only works with '--systematicType normal'",
)
parser.add_argument(
    "--histType",
    choices=["hist", "boost", "root"],
    default="hist",
    help="Type of input histogram",
)
args = parser.parse_args()

# Make histograms using different interfaces
if args.histType == "hist":
    import hist

    def Axis(*bins, axis_type="Regular", name="h1"):
        a = getattr(hist.axis, axis_type)(*bins, name=name)
        return a

    def Histogram(*axes, weighted=False):

        h = hist.Hist(
            *axes, storage=hist.storage.Weight() if weighted else hist.storage.Double()
        )
        return h

    def fill(h, *values, weight=None):
        if weight is not None:
            h.fill(*values, weight=weight)
        else:
            h.fill(*values)

elif args.histType == "boost":
    import boost_histogram as bh

    def Axis(*bins, axis_type="Regular", name=None):
        a = getattr(bh.axis, axis_type)(*bins)
        return a

    def Histogram(*axes, weighted=False):

        h = bh.Histogram(
            *axes, storage=bh.storage.Weight() if weighted else bh.storage.Double()
        )
        return h

    def fill(h, *values, weight=None):
        if weight is not None:
            h.fill(*values, weight=weight)
        else:
            h.fill(*values)

elif args.histType == "root":
    from array import array

    import ROOT

    def Axis(*bins, axis_type="Regular", name=None):
        return bins

    def Histogram(*axes, name="h2", title="Hist 1D", weighted=False):

        # format regular or variable axes
        axs = [
            (len(a[0]) - 1, array("d", a[0])) if isinstance(a[0], list) else a
            for a in axes
        ]

        if len(axes) == 1:
            h = ROOT.TH1D(name, title, *axs[0])

        elif len(axes) == 2:
            h = ROOT.TH2D(name, title, *axs[0], *axs[1])

        return h

    def fill(h, *values, weight=None):
        n = len(values[0])
        vals = [array("d", v) for v in values]
        if weight is not None:
            wgt = array("d", weight)
        else:
            wgt = array("d", [1.0] * n)
        h.FillN(n, *vals, wgt)


bins_x = [10, -5, 5]
bins_a = [10, 0, 5]
bins_b = [0, 1, 3, 6, 10, 20]

ax_x = Axis(*bins_x, axis_type="Regular", name="x")
ax_a = Axis(*bins_a, axis_type="Regular", name="a")
ax_b = Axis(bins_b, axis_type="Variable", name="b")

h1_data = Histogram(ax_x)
h2_data = Histogram(ax_a, ax_b)

h1_sig = Histogram(ax_x, weighted=True)
h2_sig = Histogram(ax_a, ax_b, weighted=True)

h1_bkg = Histogram(ax_x, weighted=True)
h2_bkg = Histogram(ax_a, ax_b, weighted=True)

h1_bkg_2 = Histogram(ax_x, weighted=True)

# masked channel e.g. for gen level distribution
h1_sig_masked = Histogram(ax_x, weighted=True)

# for pseudodata
h1_pseudo = Histogram(ax_x, weighted=True)
h2_pseudo = Histogram(ax_a, ax_b, weighted=True)

# Generate random data for filling
np.random.seed(42)  # For reproducibility


def get_sig(factor=1):
    # gaussian distributed signal
    x = np.random.normal(0, 1, 10000 * factor)
    w_x = np.random.normal(1 / factor, 0.2, 10000 * factor)

    a = np.random.normal(2, 1, 15000 * factor)
    b = np.random.normal(10, 2.5, 15000 * factor)
    w_ab = np.random.normal(1 / factor, 0.2, 15000 * factor)
    return x, w_x, a, b, w_ab


def get_sig_masked(factor=1):
    # gaussian distributed signal
    x = np.random.normal(0, 0.8, 10000 * factor)
    w_x = np.random.normal(1 / factor, 0.1, 10000 * factor)
    return x, w_x


def get_bkg(factor=1):
    # uniform distributed background
    x = np.random.uniform(-5, 5, 5000 * factor)
    w_x = np.random.normal(1 / factor, 0.2, 5000 * factor)

    a = np.random.uniform(0, 5, 7000 * factor)
    b = np.random.uniform(0, 20, 7000 * factor)
    w_ab = np.random.normal(1 / factor, 0.2, 7000 * factor)
    return x, w_x, a, b, w_ab


def get_bkg_2():
    # uniform distributed background
    x = np.random.normal(0.5, 1.5, 5000)
    return x


# Fill histograms
x, w_x, a, b, w_ab = get_sig()
fill(h1_data, x)
fill(h2_data, a, b)

x, w_x, a, b, w_ab = get_bkg()
fill(h1_data, x)
fill(h2_data, a, b)

x = get_bkg_2()
fill(h1_data, x)

x, w_x, a, b, w_ab = get_sig(3)
fill(h1_sig, x, weight=w_x)
fill(h2_sig, a, b, weight=w_ab)

x, w_x, a, b, w_ab = get_bkg(2)
fill(h1_bkg, x, weight=w_x)
fill(h2_bkg, a, b, weight=w_ab)

x = get_bkg_2()
fill(h1_bkg_2, x)

if not args.skipMaskedChannels:
    x, w_x = get_sig_masked(3)
    fill(h1_sig_masked, x, weight=w_x)

# pseudodata as exact composition of signal and background
h1_pseudo.values()[...] = (
    h1_sig.values() + h1_bkg.values()[...] + h1_bkg_2.values()[...]
)
h2_pseudo.values()[...] = h2_sig.values() + h2_bkg.values()[...]

h1_pseudo.variances()[...] = (
    h1_sig.variances() + h1_bkg.variances()[...] + h1_bkg_2.variances()[...]
)
h2_pseudo.variances()[...] = h2_sig.variances() + h2_bkg.variances()[...]

# scale signal down signal by 10%
h1_sig.values()[...] = h1_sig.values() * 0.9
h2_sig.values()[...] = h2_sig.values() * 0.9

# scale bkg up background by 5%
h1_bkg.values()[...] = h1_bkg.values() * 1.05
h2_bkg.values()[...] = h2_bkg.values() * 1.05

# scale bkg 2 down by 10%
h1_bkg_2.values()[...] = h1_bkg_2.values() * 0.9

# data covariance matrix
variances_flat = np.concatenate(
    [h1_data.values().flatten(), h2_data.values().flatten()]
)
cov = np.diag(variances_flat)

# add fully correlated contribution
variances_bkg = np.concatenate([h1_bkg.values().flatten(), h2_bkg.values().flatten()])
cov_bkg = np.diag(variances_bkg * 0.05)

# add bin by bin stat uncertainty on diagonal elements
cov += np.diag(np.concatenate([h1_sig.values().flatten(), h2_sig.values().flatten()]))
cov += np.diag(np.concatenate([h1_bkg.values().flatten(), h2_bkg.values().flatten()]))
cov += np.diag(
    np.concatenate(
        [h1_bkg_2.values().flatten(), np.zeros_like(h2_bkg.values().flatten())]
    )
)

# Build tensor
writer = tensorwriter.TensorWriter(
    sparse=args.sparse,
    systematic_type=args.systematicType,
)

writer.add_channel(h1_data.axes, "ch0")
writer.add_channel(h2_data.axes, "ch1")

writer.add_data(h1_data, "ch0")
writer.add_data(h2_data, "ch1")

writer.add_pseudodata(h1_pseudo, "original", "ch0")
writer.add_pseudodata(h2_pseudo, "original", "ch1")

writer.add_data_covariance(cov)

writer.add_process(h1_sig, "sig", "ch0", signal=True)
writer.add_process(h2_sig, "sig", "ch1", signal=True)

writer.add_process(h1_bkg, "bkg", "ch0")
writer.add_process(h2_bkg, "bkg", "ch1")

writer.add_process(h1_bkg_2, "bkg_2", "ch0")

if not args.skipMaskedChannels:
    # add masked channel
    writer.add_channel(h1_sig_masked.axes, "ch0_masked", masked=True)
    writer.add_process(h1_sig_masked, "sig", "ch0_masked", signal=True)

# systematic uncertainties

writer.add_norm_systematic("norm", ["sig", "bkg", "bkg_2"], "ch0", 1.02)
writer.add_norm_systematic("norm", ["sig", "bkg"], "ch1", [1.02, 1.03])

writer.add_norm_systematic("bkg_norm", "bkg", "ch0", 1.05)
writer.add_norm_systematic("bkg_norm", "bkg", "ch1", 1.05)

writer.add_norm_systematic("bkg_2_norm", "bkg_2", "ch0", 1.1)

# shape systematics for channel ch0

# Apply reweighting: linear function of axis value
# f(x) = a * x + b
a, b = 0.01, -0.05  # Linear coefficients
bin_centers = h1_bkg.axes[0].centers  # Get bin centers
bin_centers -= bin_centers[0]
weights = a * bin_centers + b  # Compute weights

# Reweight the histogram values
h1_bkg_syst0 = h1_bkg.copy()
h1_bkg_syst0.values()[...] = h1_bkg.values() * (1 + weights)

# unconstrained systematic that is not an NOI
writer.add_systematic(
    h1_bkg_syst0,
    "slope_background",
    "bkg",
    "ch0",
    add_to_data_covariance=args.addSystToDataCovariance,
    groups=["slopes", "slopes_background"],
    constrained=False,
)

h1_sig_syst1_up = h1_sig.copy()
h1_sig_syst1_dn = h1_sig.copy()
h1_sig_syst1_up.values()[...] = h1_sig.values() * (1 + weights)
h1_sig_syst1_dn.values()[...] = h1_sig.values() * (1 - weights)

# constrained systematic that is an NOI
writer.add_systematic(
    [h1_sig_syst1_up, h1_sig_syst1_dn],
    "slope_signal_ch0",
    "sig",
    "ch0",
    groups=["slopes", "slopes_signal"],
    symmetrize="average",
    kfactor=1.2,
    constrained=True,
    noi=True,
)

# unconstrained systematic that is an NOI
writer.add_systematic(
    [h1_sig_syst1_up, h1_sig_syst1_dn],
    "slope_signal",
    "sig",
    "ch0",
    symmetrize="average",
    constrained=False,
    noi=True,
)

if not args.skipMaskedChannels:
    h1_sig_masked_syst1_up = h1_sig_masked.copy()
    h1_sig_masked_syst1_dn = h1_sig_masked.copy()
    h1_sig_masked_syst1_up.values()[...] = h1_sig_masked.values() * (1 + weights)
    h1_sig_masked_syst1_dn.values()[...] = h1_sig_masked.values() * (1 - weights)

    writer.add_systematic(
        [h1_sig_masked_syst1_up, h1_sig_masked_syst1_dn],
        "slope_signal",
        "sig",
        "ch0_masked",
        symmetrize="average",
        constrained=False,
        noi=True,
    )

h1_sig_syst2_up = h1_sig.copy()
h1_sig_syst2_dn = h1_sig.copy()
h1_sig_syst2_up.values()[...] = h1_sig.values() * (1 + weights) ** 2
h1_sig_syst2_dn.values()[...] = h1_sig.values() * (1 - weights)

writer.add_systematic(
    [h1_sig_syst2_up, h1_sig_syst2_dn],
    "slope_lin_signal_ch0",
    "sig",
    "ch0",
    add_to_data_covariance=args.addSystToDataCovariance,
    groups=["slopes", "slopes_signal"],
    symmetrize="linear",
)

h1_sig_syst3_up = h1_sig.copy()
h1_sig_syst3_dn = h1_sig.copy()
h1_sig_syst3_up.values()[...] = h1_sig.values() * (1 + weights) ** 3
h1_sig_syst3_dn.values()[...] = h1_sig.values() * (1 - weights) ** 2

writer.add_systematic(
    [h1_sig_syst2_up, h1_sig_syst2_dn],
    "slope_quad_signal_ch0",
    "sig",
    "ch0",
    add_to_data_covariance=args.addSystToDataCovariance,
    groups=["slopes", "slopes_signal"],
    symmetrize="quadratic",
)


# shape systematics for channel ch1

bin_centers = h2_bkg.axes[0].centers  # Get bin centers
bin_centers -= bin_centers[0]
weights = (a * bin_centers + b)[..., None]  # Compute weights

h2_bkg_syst0 = h2_bkg.copy()
h2_bkg_syst0.values()[...] = h2_bkg.values() * (1 + weights)
writer.add_systematic(
    h2_bkg_syst0,
    "slope_background",
    "bkg",
    "ch1",
    add_to_data_covariance=args.addSystToDataCovariance,
    groups=["slopes", "slopes_background"],
    constrained=False,
)

h2_sig_syst1_up = h2_sig.copy()
h2_sig_syst1_dn = h2_sig.copy()
h2_sig_syst1_up.values()[...] = h2_sig.values() * (1 + weights)
h2_sig_syst1_dn.values()[...] = h2_sig.values() * (1 - weights)

writer.add_systematic(
    [h2_sig_syst1_up, h2_sig_syst1_dn],
    "slope_signal_ch1",
    "sig",
    "ch1",
    add_to_data_covariance=args.addSystToDataCovariance,
    groups=["slopes", "slopes_signal"],
    symmetrize="conservative",
)

writer.add_systematic(
    [h2_sig_syst1_up, h2_sig_syst1_dn],
    "slope_signal",
    "sig",
    "ch1",
    symmetrize="average",
    constrained=False,
    noi=True,
)

# add an asymmetric uncertainty (or symmetrize)
h2_sig_syst2_up = h2_sig.copy()
h2_sig_syst2_dn = h2_sig.copy()
h2_sig_syst2_up.values()[...] = h2_sig.values() * (1 + weights) ** 2
h2_sig_syst2_dn.values()[...] = h2_sig.values() * (1 - weights)

writer.add_systematic(
    [h2_sig_syst2_up, h2_sig_syst2_dn],
    "slope_2_signal_ch1",
    "sig",
    "ch1",
    add_to_data_covariance=args.addSystToDataCovariance,
    groups=["slopes", "slopes_signal"],
    symmetrize="quadratic" if args.symmetrizeAll else None,
)

directory = args.output
if directory == "":
    directory = "./"
filename = args.outname
if args.postfix:
    filename += f"_{args.postfix}"
writer.write(outfolder=directory, outfilename=filename)



# plot different symmetrizations

import argparse

import hist
import mplhep as hep
import numpy as np
from matplotlib.lines import Line2D

from rabbit import debugdata, inputdata, tensorwriter

from wums import output_tools, plot_tools  # isort: skip
from wums import boostHistHelpers as hh  # isort: skip


parser = argparse.ArgumentParser()
parser.add_argument("-o", "--outpath", default="./", help="output directory")
parser.add_argument(
    "--systematicType",
    choices=["log_normal", "normal"],
    default="log_normal",
    help="probability density for systematic variations",
)
parser.add_argument(
    "--title",
    default="Rabbit",
    type=str,
    help="Title to be printed in upper left",
)
parser.add_argument(
    "--subtitle",
    default="",
    type=str,
    help="Subtitle to be printed after title",
)
parser.add_argument("--titlePos", type=int, default=2, help="title position")
parser.add_argument(
    "--legPos", type=str, default="upper right", help="Set legend position"
)
parser.add_argument(
    "--legCols", type=int, default=2, help="Number of columns in legend"
)
parser.add_argument(
    "--fluctuate",
    action="store_true",
    help="To add fluctuations to the systematic histogram",
)
args = parser.parse_args()

outdir = output_tools.make_plot_dir(args.outpath)


h = hist.Hist(hist.axis.Regular(10, -0.5, 9.5, name="x"), storage=hist.storage.Double())

rng = np.random.default_rng(seed=42)
x = rng.uniform(0, 10, 100000)

h.fill(x)

# Build tensor
writer = tensorwriter.TensorWriter(
    systematic_type=args.systematicType,
)

writer.add_channel(h.axes, "ch0")

writer.add_data(h, "ch0")

writer.add_process(h, "sig", "ch0")

# systematic uncertainties

# f(x) = a * x + b
a, b = 0.01, -0.05  # Linear coefficients
bin_centers = h.axes[0].centers  # Get bin centers
weights = a * bin_centers + b  # Compute weights

a, b = 0.0005, -0.025  # Linear coefficients
weights_asym = a * bin_centers**2 + b  # Compute weights

# Reweight the histogram values
h_syst0_up = h.copy()
h_syst0_up.values()[...] = h.values() * (1 + weights)

h_syst0_down = h.copy()
h_syst0_down.values()[...] = h.values() / (1 + weights)

h_syst1_up = h.copy()
h_syst1_up.values()[...] = h.values() * (1 + weights_asym) ** 2

# add poisson fluctuations on systematics
if args.fluctuate:
    h_syst0_up.values()[...] = rng.poisson(h_syst0_up.values())
    h_syst0_down.values()[...] = rng.poisson(h_syst0_down.values())
    h_syst1_up.values()[...] = rng.poisson(h_syst1_up.values())


writer.add_systematic(
    [h_syst0_up, h_syst0_down],
    "nominal",
    "sig",
    "ch0",
    symmetrize=None,
)

writer.add_systematic(
    [h_syst0_up, h_syst0_down],
    "average",
    "sig",
    "ch0",
    symmetrize="average",
)

writer.add_systematic(
    [h_syst0_up, h_syst0_down],
    "conservative",
    "sig",
    "ch0",
    symmetrize="conservative",
)

writer.add_systematic(
    [h_syst1_up, h_syst0_down],
    "asym",
    "sig",
    "ch0",
    symmetrize=None,
)


writer.add_systematic(
    [h_syst1_up, h_syst0_down],
    "linear",
    "sig",
    "ch0",
    symmetrize="linear",
)

writer.add_systematic(
    [h_syst1_up, h_syst0_down],
    "quadratic",
    "sig",
    "ch0",
    symmetrize="quadratic",
)

# TODO: add option to skip writing and directly use object
writer.write(outfolder=outdir, outfilename="symmetrize_tensor.hdf5")

indata = inputdata.FitInputData(f"{outdir}/symmetrize_tensor.hdf5")
debug = debugdata.FitDebugData(indata)

hist_proc = debug.nominal_hists["ch0"][{"processes": "sig"}]
hist_syst = debug.syst_hists["ch0"][{"processes": "sig"}]

# check if input histograms agree with what is actually in the indata object
test_pairs = [
    (hist_proc, h, "central"),
    (hist_syst[{"DownUp": "Up", "systs": "nominal"}], h_syst0_up, "nominal-up"),
    (hist_syst[{"DownUp": "Down", "systs": "nominal"}], h_syst0_down, "nominal-down"),
]

if not args.fluctuate:
    test_pairs.extend(
        [
            (hist_syst[{"DownUp": "Up", "systs": "average"}], h_syst0_up, "average-up"),
            (
                hist_syst[{"DownUp": "Up", "systs": "conservative"}],
                h_syst0_up,
                "conservative-up",
            ),
            (
                hist_syst[{"DownUp": "Down", "systs": "average"}],
                h_syst0_down,
                "average-down",
            ),
            (
                hist_syst[{"DownUp": "Down", "systs": "conservative"}],
                h_syst0_down,
                "conservative-down",
            ),
        ]
    )

for h1, h2, stype in test_pairs:
    if not np.all(np.isclose(h1.values(), h2.values())):
        original = h1.values()
        stored = h2.values()
        raise RuntimeError(
            f"""
                        '{stype}' Histograms should agree but got
                        - original: {original}
                        - stored: {stored}
                        """
        )


labels_dict = {
    "nominal": "Asymmetric",
    "asym": "Asymmetric",
    "average": "Average",
    "conservative": "Conservative",
    "linearSymAvg": "Linear [avg.]",
    "linearSymDiff": "Linear [diff.]",
    "quadraticSymAvg": "Quadratic [avg.]",
    "quadraticSymDiff": "Quadratic [diff.]",
}

# plot variations
ratio = True  # args.systematicType == "log_normal"

xlabel = "Bin index"
if ratio:
    ylabel = "variation / nominal"
else:
    ylabel = "variation - nominal"


def make_plot(
    name, linestyles=["-", "--", ":"], stypes=["nominal", "average", "conservative"]
):

    fig, ax1 = plot_tools.figure(h, xlabel, ylabel, width_scale=0.8, ylim=[0.94, 1.09])

    # nominal hist
    hep.histplot(
        hh.divideHists(hist_proc, hist_proc),
        xerr=False,
        yerr=False,
        histtype="step",
        color="black",
        linewidth=2,
        label=None,
        density=False,
        binwnorm=None,
        ax=ax1,
        zorder=1,
        flow="none",
    )

    for i, stype in enumerate(stypes):
        hep.histplot(
            [
                (
                    hh.divideHists(hist_syst[{"DownUp": "Up", "systs": stype}], h)
                    if ratio
                    else hh.addHists(
                        hist_syst[{"DownUp": "Up", "systs": stype}], h, scale1=-1
                    )
                ),
                (
                    hh.divideHists(hist_syst[{"DownUp": "Down", "systs": stype}], h)
                    if ratio
                    else hh.addHists(
                        hist_syst[{"DownUp": "Down", "systs": stype}], h, scale1=-1
                    )
                ),
            ],
            xerr=False,
            yerr=False,
            histtype="step",
            color=["red", "blue"],
            # label=["Up", "Down"],
            linestyle=linestyles[i],
            linewidth=2,
            density=False,
            binwnorm=None,
            ax=ax1,
            zorder=1,
            flow="none",
        )

    # --- dummy handles for colors ---
    color_handles = [
        Line2D([0], [0], color="red", lw=2, label="Up"),
        Line2D([0], [0], color="blue", lw=2, label="Down"),
    ]

    # --- dummy handles for line styles ---
    style_handles = [
        Line2D([0], [0], color="black", lw=2, linestyle=ls, label=labels_dict[l])
        for ls, l in zip(linestyles, stypes)
    ]

    legend1 = ax1.legend(handles=color_handles, loc="upper left")
    legend2 = ax1.legend(handles=style_handles, loc="upper right")

    ax1.add_artist(legend1)

    plot_tools.add_decor(
        ax1,
        args.title,
        args.subtitle,
        data=True,
        lumi=None,
        loc=args.titlePos,
        no_energy=True,
    )

    plot_tools.addLegend(
        ax1,
        ncols=args.legCols,
        loc=args.legPos,
    )

    # plot_tools.fix_axes(ax1, yscale=args.yscale, logy=args.logy)

    to_join = [name, args.systematicType]
    if args.fluctuate:
        to_join.append("fluctuate")
    outfile = "_".join(filter(lambda x: x, to_join))
    if args.subtitle == "Preliminary":
        outfile += "_preliminary"

    plot_tools.save_pdf_and_png(outdir, outfile)
    analysis_meta_info = None
    output_tools.write_index_and_log(
        outdir,
        outfile,
        args=args,
    )


make_plot(
    "symmetric",
    linestyles=["-", "--", ":"],
    stypes=["nominal", "average", "conservative"],
)

make_plot(
    "linear",
    linestyles=["-", "--", ":"],
    stypes=["asym", "linearSymAvg", "linearSymDiff"],
)

make_plot(
    "quadratic",
    linestyles=["-", "--", ":"],
    stypes=["asym", "quadraticSymAvg", "quadraticSymDiff"],
)



import h5py
import hist
import numpy as np
import tensorflow as tf

from rabbit.h5pyutils import makesparsetensor, maketensor


class FitInputData:
    def __init__(self, filename, pseudodata=None):
        with h5py.File(filename, mode="r") as f:

            # load text arrays from file
            self.procs = f["hprocs"][...]
            self.signals = f["hsignals"][...]
            self.systs = f["hsysts"][...]
            self.systsnoconstraint = f["hsystsnoconstraint"][...]
            self.systgroups = f["hsystgroups"][...]
            self.systgroupidxs = f["hsystgroupidxs"][...]

            self.noiidxs = (
                f["hnoiidxs"][...]
                if "hnoiidxs" in f.keys()
                else f["hnoigroupidxs"][...]
            )

            if "hpseudodatanames" in f.keys():
                self.pseudodatanames = f["hpseudodatanames"][...].astype(str)
            else:
                self.pseudodatanames = []

            # load arrays from file

            if "hdata_cov_inv" in f.keys():
                hdata_cov_inv = f["hdata_cov_inv"]
                self.data_cov_inv = maketensor(hdata_cov_inv)
            else:
                self.data_cov_inv = None

            # load data/pseudodata
            if pseudodata is not None:
                print("Initialize pseudodata")
                hpseudodata_obs = f["hpseudodata"]

                self.pseudodata_obs = maketensor(hpseudodata_obs)
                if "hpseudodatavar" in f.keys():
                    hpseudodata_var = f["hpseudodatavar"]
                    self.pseudodata_var = maketensor(hpseudodata_var)
                else:
                    self.pseudodata_var = None

                # if explicit pseudodata sets are requested, select them (keep all for empty list)
                if len(pseudodata) > 0:
                    # Check if all pseudodata are in self.pseudodatanames
                    if not np.all(np.isin(pseudodata, self.pseudodatanames)):
                        missing = [
                            pd for pd in pseudodata if pd not in self.pseudodatanames
                        ]
                        raise ValueError(f"Missing pseudodata: {missing}")

                    mask = np.isin(self.pseudodatanames, pseudodata)
                    indices = np.where(mask)[0]
                    if not indices.size:
                        raise ValueError(
                            f"Invalid pseudodata choice {pseudodata}. Valid choices are {self.pseudodatanames}"
                        )

                    self.pseudodata_obs = tf.gather(
                        self.pseudodata_obs, indices, axis=1
                    )
                    self.pseudodatanames = self.pseudodatanames[indices]

            self.data_obs = maketensor(f["hdata_obs"])
            if "hdata_var" in f.keys():
                self.data_var = maketensor(f["hdata_var"])
            else:
                self.data_var = None

            # start by creating tensors which read in the hdf5 arrays (optimized for memory consumption)
            self.constraintweights = maketensor(f["hconstraintweights"])

            self.sparse = not "hnorm" in f

            if self.sparse:
                print(
                    "WARNING: The sparse tensor implementation is experimental and probably slower than with a dense tensor!"
                )
                self.norm = makesparsetensor(f["hnorm_sparse"])
                self.logk = makesparsetensor(f["hlogk_sparse"])
            else:
                self.norm = maketensor(f["hnorm"])
                self.logk = maketensor(f["hlogk"])

            # infer some metadata from loaded information
            self.dtype = self.data_obs.dtype
            self.nbins = self.data_obs.shape[-1]
            self.nbinsfull = self.norm.shape[0]
            self.nbinsmasked = self.nbinsfull - self.nbins
            self.nproc = len(self.procs)
            self.nsyst = len(self.systs)
            self.nsystnoconstraint = len(self.systsnoconstraint)
            self.nsignals = len(self.signals)
            self.nsystgroups = len(self.systgroups)

            # reference meta data if available
            self.metadata = {}
            if "meta" in f.keys():
                from wums.ioutils import pickle_load_h5py

                self.metadata = pickle_load_h5py(f["meta"])
                self.channel_info = self.metadata["channel_info"]
            else:
                self.channel_info = {
                    "ch0": {
                        "axes": [
                            hist.axis.Integer(
                                0,
                                self.nbins,
                                underflow=False,
                                overflow=False,
                                name="obs",
                            )
                        ]
                    }
                }
                if self.nbinsmasked:
                    self.channel_info["ch1_masked"] = {
                        "axes": [
                            hist.axis.Integer(
                                0,
                                self.nbinsmasked,
                                underflow=False,
                                overflow=False,
                                name="masked",
                            )
                        ]
                    }

            self.symmetric_tensor = self.metadata.get("symmetric_tensor", False)

            if self.metadata.get("exponential_transform", False):
                raise NotImplementedError(
                    "exponential_transform functionality has been removed.   Please use systematic_type normal instead"
                )

            self.systematic_type = self.metadata.get("systematic_type", "log_normal")

            if "hsumw2" in f.keys():
                self.sumw = maketensor(f["hsumw"])
                self.sumw2 = maketensor(f["hsumw2"])
            else:
                # fallback for older datacards
                kstat = maketensor(f["hkstat"])

                self.sumw = self.expected_events_nominal()
                self.sumw2 = self.sumw**2 / kstat

                self.sumw2 = tf.where(kstat == 0.0, self.sumw, self.sumw2)

            # compute indices for channels
            ibin = 0
            for channel, info in self.channel_info.items():
                axes = info["axes"]
                flow = info.get("flow", False)
                shape = tuple([a.extent if flow else a.size for a in axes])
                size = int(np.prod(shape))

                start = ibin
                stop = start + size

                info["start"] = start
                info["stop"] = stop

                ibin = stop

            for channel, info in self.channel_info.items():
                print(channel, info)

            self.axis_procs = hist.axis.StrCategory(self.procs, name="processes")

    @tf.function
    def expected_events_nominal(self):
        rnorm = tf.ones(self.nproc, dtype=self.dtype)
        mrnorm = tf.expand_dims(rnorm, -1)

        if self.sparse:
            nexpfullcentral = tf.sparse.sparse_dense_matmul(self.norm, mrnorm)
            nexpfullcentral = tf.squeeze(nexpfullcentral, -1)
        else:
            nexpfullcentral = tf.matmul(self.norm, mrnorm)
            nexpfullcentral = tf.squeeze(nexpfullcentral, -1)

        return nexpfullcentral



import tensorflow as tf

from .scipyhelpers import scipy_cond_number, scipy_edmval, scipy_edmval_cov


def simple_sparse_slice0end(in_sparse, end):
    """
    Slice a tf.sparse.SparseTensor along axis 0 starting 0 to the 'end'.
    """

    # Convert dense_shape, indices, and values to tensors if they aren't already
    dense_shape = in_sparse.dense_shape
    indices = in_sparse.indices
    values = in_sparse.values

    # Compute output dense shape after slicing
    out_shape = tf.concat([[end], dense_shape[1:]], axis=0)

    # Filter rows: select entries where indices[:, 0] < end
    mask = indices[:, 0] < end
    selected_indices = tf.boolean_mask(indices, mask)
    selected_values = tf.boolean_mask(values, mask)

    # Return the sliced sparse tensor
    return tf.sparse.SparseTensor(
        indices=selected_indices, values=selected_values, dense_shape=out_shape
    )


def is_diag(x):
    return tf.equal(
        tf.math.count_nonzero(x), tf.math.count_nonzero(tf.linalg.diag_part(x))
    )


def is_on_gpu(tensor):
    """
    Check if tensor is on a GPU device
    """

    device = tensor.device
    device_type = device.split(":")[-2]
    return device_type == "GPU"


def tf_edmval_cov(grad, hess):
    # use a Cholesky decomposition to easily detect the non-positive-definite case
    chol = tf.linalg.cholesky(hess)

    # FIXME catch this exception to mark failed toys and continue
    if tf.reduce_any(tf.math.is_nan(chol)).numpy():
        raise ValueError(
            "Cholesky decomposition failed, Hessian is not positive-definite"
        )

    gradv = grad[..., None]
    edmval = 0.5 * tf.linalg.matmul(
        gradv, tf.linalg.cholesky_solve(chol, gradv), transpose_a=True
    )
    edmval = edmval[0, 0].numpy()

    cov = tf.linalg.cholesky_solve(chol, tf.eye(chol.shape[0], dtype=chol.dtype))

    return edmval, cov


def edmval_cov(grad, hess):
    # scipy is faster than tensorflow on CPU so use it as appropriate
    if is_on_gpu(hess):
        return tf_edmval_cov(grad, hess)
    else:
        return scipy_edmval_cov(grad.__array__(), hess.__array__())


def tf_edmval(grad, hess):
    # Ensure proper shapes
    grad = tf.reshape(grad, (-1, 1))  # shape (n, 1)

    # Solve H x = g for x
    x = tf.linalg.solve(hess, grad)  # shape (n, 1)

    # Compute EDM = 0.5 * g^T x
    edm = 0.5 * tf.squeeze(tf.matmul(tf.transpose(grad), x))

    return edmval


def edmval(grad, hess):
    # scipy is faster than tensorflow on CPU so use it as appropriate
    if is_on_gpu(hess):
        return tf_edmval(grad, hess)
    else:
        return scipy_edmval(grad.__array__(), hess.__array__())


def cond_number(hess):
    if is_on_gpu(hess):
        return tf.linalg.cond(hess)
    else:
        return scipy_cond_number(hess.__array__())


def segment_sum_along_axis(x, segment_ids, idx, num_segments):
    # Move the target axis to the front
    perm = [idx] + [i for i in range(len(x.shape)) if i != idx]
    x_transposed = tf.transpose(x, perm)

    # Count leading/trailing -1s and cut them off
    nleading = tf.reduce_sum(
        tf.cast(
            tf.cumsum(tf.cast(segment_ids == -1, tf.int32))
            == tf.range(1, len(segment_ids) + 1),
            tf.int32,
        )
    )
    ntrailing = tf.reduce_sum(
        tf.cast(
            tf.cumsum(tf.cast(segment_ids[::-1] == -1, tf.int32))
            == tf.range(1, len(segment_ids) + 1),
            tf.int32,
        )
    )
    x_transposed = x_transposed[nleading : len(x_transposed) - ntrailing]
    segment_ids_valid = segment_ids[nleading : len(segment_ids) - ntrailing]

    # Apply segment_sum along axis 0
    rebinned = tf.math.segment_sum(x_transposed, segment_ids_valid)

    # Update static shape if possible
    static_shape = rebinned.shape.as_list()
    static_shape[0] = num_segments
    rebinned.set_shape(static_shape)

    # Undo the transposition
    reverse_perm = [perm.index(i) for i in range(len(perm))]
    rebinned = tf.transpose(rebinned, reverse_perm)

    return rebinned
